# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import asynccontextmanager
from pathlib import Path
from typing import (
    AsyncGenerator,
    Callable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)
from unittest.mock import MagicMock, patch

import httpx
import pytest

from fastapi import Request

from kserve.protocol.rest.openai import (
    ChatPrompt,
    OpenAIChatAdapterModel,
    OpenAIProxyModel,
)

from kserve.protocol.rest.openai.types import (
    Completion,
    CompletionRequest,
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionRequest,
    Error,
    ErrorResponse,
)
from kserve.protocol.rest.openai.errors import OpenAIError

FIXTURES_PATH = Path(__file__).parent / "fixtures" / "openai"


class ChunkIterator:
    """Yields chunks"""

    def __init__(self, chunks: List[Completion]):
        self.chunks = chunks
        self.curr_chunk = 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.curr_chunk >= len(self.chunks):
            raise StopAsyncIteration
        chunk = self.chunks[self.curr_chunk]
        self.curr_chunk += 1
        return chunk


class DummyModel(OpenAIChatAdapterModel):
    data: Tuple[Completion, Completion]
    num_chunks: int

    def __init__(self, data: Tuple[Completion, Completion], num_chunks: int = 5):
        self.data = data
        self.num_chunks = num_chunks

    async def create_completion(
        self,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
    ) -> Union[AsyncGenerator[str, None], Completion, ErrorResponse]:
        if request.stream:

            async def stream_results() -> AsyncGenerator[str, None]:
                async for partial_completion in ChunkIterator(
                    [self.data[1]] * self.num_chunks
                ):
                    yield f"data: {partial_completion.model_dump_json()}\n\n"
                yield "data: [DONE]\n\n"

            return stream_results()
        else:
            return self.data[0]

    def apply_chat_template(
        self,
        request: ChatCompletionRequest,
    ) -> ChatPrompt:
        return ChatPrompt(prompt="hello")


@pytest.fixture
def completion():
    with open(FIXTURES_PATH / "completion.json") as f:
        return Completion.model_validate_json(f.read())


@pytest.fixture
def completion_partial():
    with open(FIXTURES_PATH / "completion_partial.json") as f:
        return Completion.model_validate_json(f.read())


@pytest.fixture
def completion_chunk_stream():
    with open(FIXTURES_PATH / "completion_chunk_stream.txt", "rb") as f:
        return f.read()


@pytest.fixture
def chat_completion():
    with open(FIXTURES_PATH / "chat_completion.json") as f:
        return ChatCompletion.model_validate_json(f.read())


@pytest.fixture
def chat_completion_chunk():
    with open(FIXTURES_PATH / "chat_completion_chunk.json") as f:
        return ChatCompletionChunk.model_validate_json(f.read())


@pytest.fixture
def chat_completion_chunk_stream():
    with open(FIXTURES_PATH / "chat_completion_chunk_stream.txt", "rb") as f:
        return f.read()


@pytest.fixture
def completion_create_params():
    with open(FIXTURES_PATH / "completion_create_params.json") as f:
        return CompletionRequest.model_validate_json(f.read())


@pytest.fixture
def chat_completion_create_params():
    with open(FIXTURES_PATH / "chat_completion_create_params.json") as f:
        return ChatCompletionRequest.model_validate_json(f.read())


@pytest.fixture
def completion_request(completion_create_params: CompletionRequest):
    return completion_create_params


@pytest.fixture
def chat_completion_request(chat_completion_create_params: ChatCompletionRequest):
    return chat_completion_create_params


@pytest.fixture
def dummy_model(completion: Completion, completion_partial: Completion):
    return DummyModel((completion, completion_partial))


@asynccontextmanager
async def mocked_openai_proxy_model(handler: Callable):
    transport = httpx.MockTransport(handler=handler)
    http_client = httpx.AsyncClient(transport=transport)
    try:
        with patch.object(
            OpenAIProxyModel, "preprocess_completion_request"
        ), patch.object(OpenAIProxyModel, "postprocess_completion"), patch.object(
            OpenAIProxyModel, "postprocess_completion_chunk"
        ), patch.object(
            OpenAIProxyModel, "preprocess_chat_completion_request"
        ), patch.object(
            OpenAIProxyModel, "postprocess_chat_completion"
        ), patch.object(
            OpenAIProxyModel, "postprocess_chat_completion_chunk"
        ):
            yield OpenAIProxyModel(
                name="test-model",
                predictor_url="http://example.com/",
                http_client=http_client,
            )
    finally:
        await http_client.aclose()


class TestOpenAICreateCompletion:
    def test_completion_to_chat_completion(
        self, completion: Completion, chat_completion: ChatCompletion
    ):
        converted_chat_completion = (
            OpenAIChatAdapterModel.completion_to_chat_completion(
                completion, "assistant"
            )
        )
        assert (
            converted_chat_completion.model_dump_json()
            == chat_completion.model_dump_json()
        )

    @pytest.mark.asyncio
    async def test_create_completion_not_streaming(
        self,
        dummy_model: DummyModel,
        completion: Completion,
        completion_create_params: CompletionRequest,
    ):
        c = await dummy_model.create_completion(completion_create_params)
        assert isinstance(c, Completion)
        assert c.model_dump_json(indent=2) == completion.model_dump_json(indent=2)

    @pytest.mark.asyncio
    async def test_create_completion_streaming(
        self,
        dummy_model: DummyModel,
        completion_partial: Completion,
        completion_create_params: CompletionRequest,
    ):
        completion_create_params.stream = True
        c = await dummy_model.create_completion(completion_create_params)
        assert isinstance(c, AsyncGenerator)
        num_chunks_consumed = 0
        async for chunk in c:
            chunk = chunk.removeprefix("data: ")
            if chunk == "[DONE]\n\n":
                return
            assert (
                chunk == completion_partial.model_dump_json() + "\n\n"
            )  # not using indent but using two new lines
            num_chunks_consumed += 1
        assert num_chunks_consumed == dummy_model.num_chunks


class TestOpenAICreateChatCompletion:
    def test_completion_to_chat_completion(
        self, completion: Completion, chat_completion: ChatCompletion
    ):
        converted_chat_completion = (
            OpenAIChatAdapterModel.completion_to_chat_completion(
                completion, "assistant"
            )
        )
        assert (
            converted_chat_completion.model_dump_json()
            == chat_completion.model_dump_json()
        )

    @pytest.mark.asyncio
    async def test_create_chat_completion_not_streaming(
        self,
        dummy_model: DummyModel,
        chat_completion: ChatCompletion,
        chat_completion_create_params: ChatCompletionRequest,
    ):
        c = await dummy_model.create_chat_completion(chat_completion_create_params)
        assert isinstance(c, ChatCompletion)
        assert c.model_dump_json(indent=2) == chat_completion.model_dump_json(indent=2)

    @pytest.mark.asyncio
    async def test_create_chat_completion_streaming(
        self,
        dummy_model: DummyModel,
        chat_completion_chunk: ChatCompletionChunk,
        chat_completion_create_params: ChatCompletionRequest,
    ):
        chat_completion_create_params.stream = True
        c = await dummy_model.create_chat_completion(chat_completion_create_params)
        assert isinstance(c, AsyncGenerator)
        num_chunks_consumed = 0
        async for chunk in c:
            chunk = chunk.removeprefix("data: ")
            if chunk == "[DONE]\n\n":
                return
            assert (
                chunk == chat_completion_chunk.model_dump_json() + "\n\n"
            )  # not using indent but using two new lines
            num_chunks_consumed += 1
        assert num_chunks_consumed == dummy_model.num_chunks


class TestOpenAICompletionConversion:
    def test_completion_to_chat_completion(
        self, completion: Completion, chat_completion: ChatCompletion
    ):
        converted_chat_completion = (
            OpenAIChatAdapterModel.completion_to_chat_completion(
                completion, "assistant"
            )
        )
        assert (
            converted_chat_completion.model_dump_json()
            == chat_completion.model_dump_json()
        )

    def test_completion_to_chat_completion_chunk(
        self, completion_partial: Completion, chat_completion_chunk: ChatCompletionChunk
    ):
        converted_chat_completion_chunk = (
            OpenAIChatAdapterModel.completion_to_chat_completion_chunk(
                completion_partial, "assistant"
            )
        )
        assert converted_chat_completion_chunk.model_dump_json(
            indent=2
        ) == chat_completion_chunk.model_dump_json(indent=2)


class TestOpenAIParamsConversion:
    def test_convert_params(
        self,
        chat_completion_create_params: ChatCompletionRequest,
        completion_create_params: CompletionRequest,
    ):
        converted_params = (
            OpenAIChatAdapterModel.chat_completion_params_to_completion_params(
                chat_completion_create_params,
                prompt=chat_completion_create_params.messages[0]["content"],
            )
        )
        assert converted_params == completion_create_params


class TestOpenAIProxyModelCompletion:

    @pytest.mark.asyncio
    async def test_completion_upstream_connection_error(
        self,
        completion_request: CompletionRequest,
        completion: Completion,
    ):
        def handler(request):
            raise httpx.ConnectError(
                "[Errno 8] nodename nor servname provided, or not known"
            )

        async with mocked_openai_proxy_model(handler) as model:
            with pytest.raises(OpenAIError) as err_info:
                result = await model.create_completion(completion_request)
                assert result == completion
        assert (
            "Failed to communicate with upstream: [Errno 8] nodename nor servname provided"
            in str(err_info.value)
        )

    @pytest.mark.asyncio
    async def test_completion_upstream_status_code_error_invalid_body(
        self,
        completion_request: CompletionRequest,
        completion: Completion,
    ):
        def handler(request):
            return httpx.Response(status_code=400, content="Junk response")

        async with mocked_openai_proxy_model(handler) as model:
            with pytest.raises(OpenAIError) as err_info:
                result = await model.create_completion(completion_request)
                assert result == completion
        assert "Received invalid response from upstream: Junk response" in str(
            err_info.value
        )

    @pytest.mark.asyncio
    async def test_completion_upstream_status_code_error_valid_body(
        self,
        completion_request: CompletionRequest,
        completion: Completion,
    ):
        res = ErrorResponse(
            error=Error(
                code="400", message="Bad request", type="BadRequest", param=None
            )
        )

        def handler(request):
            return httpx.Response(status_code=400, content=res.model_dump_json())

        async with mocked_openai_proxy_model(handler) as model:
            with pytest.raises(OpenAIError) as err_info:
                result = await model.create_completion(completion_request)
                assert result == completion
        assert err_info.value.response == res

    @pytest.mark.asyncio
    async def test_completion_upstream_timeout(
        self,
        completion_request: CompletionRequest,
        completion: Completion,
    ):
        def handler(request):
            raise httpx.ReadTimeout("Read timed out", request=request)

        async with mocked_openai_proxy_model(handler) as model:
            with pytest.raises(OpenAIError) as err_info:
                result = await model.create_completion(completion_request)
                assert result == completion
        assert (
            str(err_info.value)
            == "Timed out when communicating with upstream: Read timed out"
        )

    @pytest.mark.asyncio
    async def test_completion_upstream_request_error(
        self,
        completion_request: CompletionRequest,
        completion: Completion,
    ):
        def handler(request):
            raise httpx.RequestError("Some error", request=request)

        async with mocked_openai_proxy_model(handler) as model:
            with pytest.raises(OpenAIError) as err_info:
                result = await model.create_completion(completion_request)
                assert result == completion
        assert str(err_info.value) == "Upstream request failed: Some error"

    @pytest.mark.asyncio
    async def test_completion_non_streamed(
        self,
        completion_request: CompletionRequest,
        completion: Completion,
    ):
        def handler(request):
            return httpx.Response(
                200,
                headers={"Content-type": "application/json"},
                content=completion.model_dump_json(),
            )

        async with mocked_openai_proxy_model(handler) as model:
            result = await model.create_completion(completion_request)
            assert result == completion
            cast(
                MagicMock, OpenAIProxyModel.preprocess_completion_request
            ).assert_called_once_with(completion_request, None)
            cast(
                MagicMock, OpenAIProxyModel.postprocess_completion
            ).assert_called_once_with(completion, completion_request, None)
            cast(
                MagicMock, OpenAIProxyModel.preprocess_chat_completion_request
            ).assert_not_called()
            cast(
                MagicMock, OpenAIProxyModel.postprocess_chat_completion
            ).assert_not_called()

    @pytest.mark.asyncio
    async def test_completion_streamed(
        self,
        completion_request: CompletionRequest,
        completion_chunk_stream: bytes,
    ):
        completion_request.stream = True

        def handler(request):
            return httpx.Response(
                200,
                headers={"Content-type": "text/event-stream"},
                content=completion_chunk_stream,
            )

        async with mocked_openai_proxy_model(handler) as model:
            async for completion_chunk in await model.create_completion(
                completion_request
            ):
                cast(
                    MagicMock, OpenAIProxyModel.postprocess_completion_chunk
                ).assert_called_with(completion_chunk, completion_request, None)
            cast(
                MagicMock, OpenAIProxyModel.preprocess_completion_request
            ).assert_called_once_with(completion_request, None)
            cast(MagicMock, OpenAIProxyModel.postprocess_completion).assert_not_called()
            cast(
                MagicMock, OpenAIProxyModel.preprocess_chat_completion_request
            ).assert_not_called()
            cast(
                MagicMock, OpenAIProxyModel.postprocess_chat_completion
            ).assert_not_called()

    @pytest.mark.asyncio
    async def test_chat_completion_non_streamed(
        self,
        chat_completion_request: ChatCompletionRequest,
        chat_completion: ChatCompletion,
    ):
        def handler(request):
            return httpx.Response(
                200,
                headers={"Content-type": "application/json"},
                content=chat_completion.model_dump_json(),
            )

        async with mocked_openai_proxy_model(handler) as model:
            result = await model.create_chat_completion(chat_completion_request)
            assert result == chat_completion
            cast(
                MagicMock, OpenAIProxyModel.preprocess_chat_completion_request
            ).assert_called_once_with(chat_completion_request, None)
            cast(
                MagicMock, OpenAIProxyModel.postprocess_chat_completion
            ).assert_called_once_with(chat_completion, chat_completion_request, None)
            cast(
                MagicMock, OpenAIProxyModel.preprocess_completion_request
            ).assert_not_called()
            cast(MagicMock, OpenAIProxyModel.postprocess_completion).assert_not_called()

    @pytest.mark.asyncio
    async def test_chat_completion_streamed(
        self,
        chat_completion_request: ChatCompletionRequest,
        chat_completion_chunk_stream: bytes,
    ):
        chat_completion_request.stream = True

        def handler(request):
            return httpx.Response(
                200,
                headers={"Content-type": "text/event-stream"},
                content=chat_completion_chunk_stream,
            )

        async with mocked_openai_proxy_model(handler) as model:
            async for chat_completion_chunk in await model.create_chat_completion(
                chat_completion_request
            ):
                cast(
                    MagicMock, OpenAIProxyModel.postprocess_chat_completion_chunk
                ).assert_called_with(
                    chat_completion_chunk, chat_completion_request, None
                )
            cast(
                MagicMock, OpenAIProxyModel.preprocess_chat_completion_request
            ).assert_called_once_with(chat_completion_request, None)
            cast(MagicMock, OpenAIProxyModel.postprocess_completion).assert_not_called()
            cast(
                MagicMock, OpenAIProxyModel.preprocess_completion_request
            ).assert_not_called()
            cast(
                MagicMock, OpenAIProxyModel.postprocess_chat_completion
            ).assert_not_called()

    @pytest.mark.asyncio
    async def test_healthcheck_success(self):
        def handler(request):
            if request.url.path == "/health":
                return httpx.Response(
                    200,
                )
            return httpx.Response(
                503,
            )

        async with mocked_openai_proxy_model(handler) as model:
            result = await model.healthy()
            assert result is True

    @pytest.mark.asyncio
    async def test_healthcheck_failure(self):
        def handler(request):
            if request.url.path == "/health":
                return httpx.Response(
                    503,
                )
            return httpx.Response(
                200,
            )

        async with mocked_openai_proxy_model(handler) as model:
            result = await model.healthy()
            assert result is False



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, wraps
from http import HTTPStatus
from typing import Any, Dict, AsyncGenerator, Optional, Union
from fastapi import Request

import httpx
import orjson
from pydantic import ValidationError

from ....logging import logger
from .errors import OpenAIError, create_error_response
from .openai_model import (
    OpenAIGenerativeModel,
    AsyncMappingIterator,
    CompletionRequest,
    ChatCompletionRequest,
)
from .types import (
    ChatCompletion,
    ChatCompletionChunk,
    Completion,
    CompletionChunk,
    ErrorResponse,
)

COMPLETIONS_ENDPOINT = "/v1/completions"
CHAT_COMPLETIONS_ENDPOINT = "/v1/chat/completions"


def error_handler(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            res = await f(*args, **kwargs)
            return res
        except httpx.HTTPStatusError as e:
            try:
                # Try to parse upstream error as an ErrorResponse object
                response = ErrorResponse.model_validate_json(e.response.content)
            except ValidationError:
                logger.warning(
                    f"Failed to parse error response from upstream: {e.response.content}"
                )
                response = create_error_response(
                    f"Received invalid response from upstream: {e.response.text}",
                    status_code=HTTPStatus.BAD_GATEWAY,
                    err_type="BadGateway",
                )
            raise OpenAIError(response=response)

        except httpx.TimeoutException as e:
            raise OpenAIError(
                response=create_error_response(
                    f"Timed out when communicating with upstream: {e}",
                    err_type="GatewayTimeout",
                    status_code=HTTPStatus.GATEWAY_TIMEOUT,
                )
            )
        except httpx.NetworkError as e:
            raise OpenAIError(
                response=create_error_response(
                    f"Failed to communicate with upstream: {e}",
                    err_type="ServiceUnavailableError",
                    status_code=HTTPStatus.SERVICE_UNAVAILABLE,
                )
            )
        except httpx.HTTPError as e:
            raise OpenAIError(
                response=create_error_response(
                    f"Upstream request failed: {e}",
                    err_type="InternalServerError",
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                )
            )

    return wrapper


class OpenAIProxyModel(OpenAIGenerativeModel):
    """
    An implementation of OpenAIModel that proxies requests to a backend server exposing Open AI endpoints.

    Users can extend this class and override any of the following methods to hook into the request/response cycle:
        - preprocess_completion_request
        - postprocess_completion
        - postprocess_completion_chunk
        - preprocess_chat_completion_request
        - postprocess_chat_completion
        - postprocess_chat_completion_chunk

    Each method takes a single parameter: the object currently being processed. Request objects may be mutated to modify
    the request sent to the downstream server and response objects (completions/completion chunks) may be mutated to
    modify the response returned to the upstream caller.

    Args:
        predictor_url (str):
            The url of the model server to send requests to. Should be fully qualified with scheme, host, and port.
            e.g. `http://my-backend:9000`
        http_client (httpx.AsyncClient|None):
            An optional instance of httpx.AsyncClient to use for sending requests to the upstream server.
        health_endpoint: The default /health endpoint is for TGI and vLLM, use /openai/v1/models/{model_name}
            for KServe OpenAI protocol.
    """

    predictor_url: str
    skip_upstream_validation: bool
    _http_client: httpx.AsyncClient
    _completions_endpoint: str
    _chat_completions_endpoint: str
    _health_endpoint: Optional[str] = None

    def __init__(
        self,
        name: str,
        predictor_url: str,
        http_client: Optional[httpx.AsyncClient] = None,
        skip_upstream_validation: bool = False,
        health_endpoint: Optional[str] = "/health",
    ):
        super().__init__(name)
        self.predictor_url = predictor_url
        self._http_client = (
            http_client
            if http_client is not None
            else httpx.AsyncClient(timeout=httpx.Timeout(timeout=5.0, read=30.0))
        )
        self._completions_endpoint = (
            f"{self.predictor_url.rstrip('/')}{COMPLETIONS_ENDPOINT}"
        )
        self._chat_completions_endpoint = (
            f"{self.predictor_url.rstrip('/')}{CHAT_COMPLETIONS_ENDPOINT}"
        )
        if health_endpoint:
            self._health_endpoint = f"{self.predictor_url.rstrip('/')}{health_endpoint}"
        self.skip_upstream_validation = skip_upstream_validation
        self.ready = True

    def preprocess_completion_request(
        self,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        """Preprocess a completion request."""
        pass

    def postprocess_completion(
        self,
        completion: Completion,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        """Postprocess a completion. Only called when response is not being streamed (i.e. stream=false)"""
        pass

    def postprocess_completion_chunk(
        self,
        completion: Completion,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        """Postprocess a completion chunk. Only called when response is being streamed (i.e. stream=true)
        This method will be called once for each chunk that is streamed back to the user.
        """
        pass

    def preprocess_chat_completion_request(
        self,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        """Preprocess a chat completion request."""
        pass

    def postprocess_chat_completion(
        self,
        chat_completion: ChatCompletion,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        """Postprocess a chat completion. Only called when response is not being streamed (i.e. stream=false)"""
        pass

    def postprocess_chat_completion_chunk(
        self,
        chat_completion_chunk: ChatCompletionChunk,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        """Postprocess a chat completion chunk. Only called when response is being streamed (i.e. stream=true)
        This method will be called once for each chunk that is streamed back to the user.
        """
        pass

    def _handle_completion_chunk(
        self,
        raw_chunk: str,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        # Skip empty lines
        if len(raw_chunk) == 0:
            return None
        # All chunks are prefixed with "data: "
        data = raw_chunk[6:]
        if data == "[DONE]":
            return None

        if self.skip_upstream_validation:
            obj = orjson.loads(data)
            completion_chunk = CompletionChunk.model_construct(**obj)
        else:
            completion_chunk = CompletionChunk.model_validate_json(data)
        self.postprocess_completion_chunk(completion_chunk, request, raw_request)
        return completion_chunk

    def _handle_chat_completion_chunk(
        self,
        raw_chunk: str,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
    ):
        # Skip empty lines
        if len(raw_chunk) == 0:
            return None
        # All chunks are prefixed with "data: "
        if len(raw_chunk) == 0:
            return None
        data = raw_chunk[6:]
        if data == "[DONE]":
            return None
        if self.skip_upstream_validation:
            obj = orjson.loads(data)
            chat_completion_chunk = ChatCompletionChunk.model_construct(**obj)
        else:
            chat_completion_chunk = ChatCompletionChunk.model_validate_json(data)
        self.postprocess_chat_completion_chunk(
            chat_completion_chunk, request, raw_request
        )
        return chat_completion_chunk

    def _build_request(
        self,
        endpoint: str,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
    ) -> httpx.Request:

        if raw_request and "upstream_headers" in raw_request:
            headers = httpx.Headers(raw_request["upstream_headers"])
        else:
            headers = httpx.Headers()

        headers["Content-type"] = "application/json"

        req = self._http_client.build_request(
            "POST",
            endpoint,
            content=request.model_dump_json(exclude_unset=True, exclude_none=True),
            headers=headers,
        )
        return req

    @error_handler
    async def create_completion(
        self,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], Completion, ErrorResponse]:
        self.preprocess_completion_request(request, raw_request)
        if request.stream:
            req = self._build_request(self._completions_endpoint, request, raw_request)
            r = await self._http_client.send(req, stream=True)
            r.raise_for_status()
            it = AsyncMappingIterator(
                iterator=r.aiter_lines(),
                mapper=partial(
                    self._handle_completion_chunk,
                    request=request,
                    raw_request=raw_request,
                ),
                close=r.aclose,
            )
            return it
        else:
            completion = await self.generate_completion(request, raw_request)
            self.postprocess_completion(completion, request, raw_request)
            return completion

    async def generate_completion(
        self, request: CompletionRequest, raw_request: Optional[Request] = None
    ) -> Completion:
        req = self._build_request(self._completions_endpoint, request, raw_request)
        response = await self._http_client.send(req)
        response.raise_for_status()
        if self.skip_upstream_validation:
            obj = response.json()
            completion = Completion.model_construct(**obj)
        else:
            completion = Completion.model_validate_json(response.content)
        return completion

    @error_handler
    async def create_chat_completion(
        self,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], ChatCompletion, ErrorResponse]:
        self.preprocess_chat_completion_request(request, raw_request)
        if request.stream:
            req = self._build_request(
                self._chat_completions_endpoint, request, raw_request
            )
            r = await self._http_client.send(req, stream=True)
            r.raise_for_status()
            it = AsyncMappingIterator(
                iterator=r.aiter_lines(),
                mapper=partial(
                    self._handle_chat_completion_chunk,
                    request=request,
                    raw_request=raw_request,
                ),
                close=r.aclose,
            )
            return it
        else:
            chat_completion = await self.generate_chat_completion(request, raw_request)
            self.postprocess_chat_completion(chat_completion, request, raw_request)
            return chat_completion

    async def generate_chat_completion(
        self,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
    ) -> ChatCompletion:
        req = self._build_request(self._chat_completions_endpoint, request, raw_request)
        response = await self._http_client.send(req)
        response.raise_for_status()
        if self.skip_upstream_validation:
            obj = response.json()
            chat_completion = ChatCompletion.model_construct(**obj)
        else:
            chat_completion = ChatCompletion.model_validate_json(response.content)
        return chat_completion

    async def healthy(self) -> bool:
        if not self._health_endpoint:
            return super().healthy()
        req = self._http_client.build_request(
            "GET",
            self._health_endpoint,
        )
        response = await self._http_client.send(req)
        return response.status_code == httpx.codes.OK



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1alpha1InferenceGraphSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'affinity': 'V1Affinity',
        'max_replicas': 'int',
        'min_replicas': 'int',
        'node_name': 'str',
        'node_selector': 'dict(str, str)',
        'nodes': 'dict(str, V1alpha1InferenceRouter)',
        'resources': 'V1ResourceRequirements',
        'router_timeouts': 'V1alpha1InfereceGraphRouterTimeouts',
        'scale_metric': 'str',
        'scale_target': 'int',
        'service_account_name': 'str',
        'timeout': 'int',
        'tolerations': 'list[V1Toleration]'
    }

    attribute_map = {
        'affinity': 'affinity',
        'max_replicas': 'maxReplicas',
        'min_replicas': 'minReplicas',
        'node_name': 'nodeName',
        'node_selector': 'nodeSelector',
        'nodes': 'nodes',
        'resources': 'resources',
        'router_timeouts': 'routerTimeouts',
        'scale_metric': 'scaleMetric',
        'scale_target': 'scaleTarget',
        'service_account_name': 'serviceAccountName',
        'timeout': 'timeout',
        'tolerations': 'tolerations'
    }

    def __init__(self, affinity=None, max_replicas=None, min_replicas=None, node_name=None, node_selector=None, nodes=None, resources=None, router_timeouts=None, scale_metric=None, scale_target=None, service_account_name=None, timeout=None, tolerations=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1InferenceGraphSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._affinity = None
        self._max_replicas = None
        self._min_replicas = None
        self._node_name = None
        self._node_selector = None
        self._nodes = None
        self._resources = None
        self._router_timeouts = None
        self._scale_metric = None
        self._scale_target = None
        self._service_account_name = None
        self._timeout = None
        self._tolerations = None
        self.discriminator = None

        if affinity is not None:
            self.affinity = affinity
        if max_replicas is not None:
            self.max_replicas = max_replicas
        if min_replicas is not None:
            self.min_replicas = min_replicas
        if node_name is not None:
            self.node_name = node_name
        if node_selector is not None:
            self.node_selector = node_selector
        self.nodes = nodes
        if resources is not None:
            self.resources = resources
        if router_timeouts is not None:
            self.router_timeouts = router_timeouts
        if scale_metric is not None:
            self.scale_metric = scale_metric
        if scale_target is not None:
            self.scale_target = scale_target
        if service_account_name is not None:
            self.service_account_name = service_account_name
        if timeout is not None:
            self.timeout = timeout
        if tolerations is not None:
            self.tolerations = tolerations

    @property
    def affinity(self):
        """Gets the affinity of this V1alpha1InferenceGraphSpec.  # noqa: E501


        :return: The affinity of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: V1Affinity
        """
        return self._affinity

    @affinity.setter
    def affinity(self, affinity):
        """Sets the affinity of this V1alpha1InferenceGraphSpec.


        :param affinity: The affinity of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: V1Affinity
        """

        self._affinity = affinity

    @property
    def max_replicas(self):
        """Gets the max_replicas of this V1alpha1InferenceGraphSpec.  # noqa: E501

        Maximum number of replicas for autoscaling.  # noqa: E501

        :return: The max_replicas of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: int
        """
        return self._max_replicas

    @max_replicas.setter
    def max_replicas(self, max_replicas):
        """Sets the max_replicas of this V1alpha1InferenceGraphSpec.

        Maximum number of replicas for autoscaling.  # noqa: E501

        :param max_replicas: The max_replicas of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: int
        """

        self._max_replicas = max_replicas

    @property
    def min_replicas(self):
        """Gets the min_replicas of this V1alpha1InferenceGraphSpec.  # noqa: E501

        Minimum number of replicas, defaults to 1 but can be set to 0 to enable scale-to-zero.  # noqa: E501

        :return: The min_replicas of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: int
        """
        return self._min_replicas

    @min_replicas.setter
    def min_replicas(self, min_replicas):
        """Sets the min_replicas of this V1alpha1InferenceGraphSpec.

        Minimum number of replicas, defaults to 1 but can be set to 0 to enable scale-to-zero.  # noqa: E501

        :param min_replicas: The min_replicas of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: int
        """

        self._min_replicas = min_replicas

    @property
    def node_name(self):
        """Gets the node_name of this V1alpha1InferenceGraphSpec.  # noqa: E501

        NodeName specifies the node name for the InferenceGraph. https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/  # noqa: E501

        :return: The node_name of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: str
        """
        return self._node_name

    @node_name.setter
    def node_name(self, node_name):
        """Sets the node_name of this V1alpha1InferenceGraphSpec.

        NodeName specifies the node name for the InferenceGraph. https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/  # noqa: E501

        :param node_name: The node_name of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: str
        """

        self._node_name = node_name

    @property
    def node_selector(self):
        """Gets the node_selector of this V1alpha1InferenceGraphSpec.  # noqa: E501

        NodeSelector specifies the node selector for the InferenceGraph. https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/  # noqa: E501

        :return: The node_selector of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._node_selector

    @node_selector.setter
    def node_selector(self, node_selector):
        """Sets the node_selector of this V1alpha1InferenceGraphSpec.

        NodeSelector specifies the node selector for the InferenceGraph. https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/  # noqa: E501

        :param node_selector: The node_selector of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: dict(str, str)
        """

        self._node_selector = node_selector

    @property
    def nodes(self):
        """Gets the nodes of this V1alpha1InferenceGraphSpec.  # noqa: E501

        Map of InferenceGraph router nodes Each node defines the router which can be different routing types  # noqa: E501

        :return: The nodes of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: dict(str, V1alpha1InferenceRouter)
        """
        return self._nodes

    @nodes.setter
    def nodes(self, nodes):
        """Sets the nodes of this V1alpha1InferenceGraphSpec.

        Map of InferenceGraph router nodes Each node defines the router which can be different routing types  # noqa: E501

        :param nodes: The nodes of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: dict(str, V1alpha1InferenceRouter)
        """
        if self.local_vars_configuration.client_side_validation and nodes is None:  # noqa: E501
            raise ValueError("Invalid value for `nodes`, must not be `None`")  # noqa: E501

        self._nodes = nodes

    @property
    def resources(self):
        """Gets the resources of this V1alpha1InferenceGraphSpec.  # noqa: E501


        :return: The resources of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: V1ResourceRequirements
        """
        return self._resources

    @resources.setter
    def resources(self, resources):
        """Sets the resources of this V1alpha1InferenceGraphSpec.


        :param resources: The resources of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: V1ResourceRequirements
        """

        self._resources = resources

    @property
    def router_timeouts(self):
        """Gets the router_timeouts of this V1alpha1InferenceGraphSpec.  # noqa: E501


        :return: The router_timeouts of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: V1alpha1InfereceGraphRouterTimeouts
        """
        return self._router_timeouts

    @router_timeouts.setter
    def router_timeouts(self, router_timeouts):
        """Sets the router_timeouts of this V1alpha1InferenceGraphSpec.


        :param router_timeouts: The router_timeouts of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: V1alpha1InfereceGraphRouterTimeouts
        """

        self._router_timeouts = router_timeouts

    @property
    def scale_metric(self):
        """Gets the scale_metric of this V1alpha1InferenceGraphSpec.  # noqa: E501

        ScaleMetric defines the scaling metric type watched by autoscaler possible values are concurrency, rps, cpu, memory. concurrency, rps are supported via Knative Pod Autoscaler(https://knative.dev/docs/serving/autoscaling/autoscaling-metrics).  # noqa: E501

        :return: The scale_metric of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: str
        """
        return self._scale_metric

    @scale_metric.setter
    def scale_metric(self, scale_metric):
        """Sets the scale_metric of this V1alpha1InferenceGraphSpec.

        ScaleMetric defines the scaling metric type watched by autoscaler possible values are concurrency, rps, cpu, memory. concurrency, rps are supported via Knative Pod Autoscaler(https://knative.dev/docs/serving/autoscaling/autoscaling-metrics).  # noqa: E501

        :param scale_metric: The scale_metric of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: str
        """

        self._scale_metric = scale_metric

    @property
    def scale_target(self):
        """Gets the scale_target of this V1alpha1InferenceGraphSpec.  # noqa: E501

        ScaleTarget specifies the integer target value of the metric type the Autoscaler watches for. concurrency and rps targets are supported by Knative Pod Autoscaler (https://knative.dev/docs/serving/autoscaling/autoscaling-targets/).  # noqa: E501

        :return: The scale_target of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: int
        """
        return self._scale_target

    @scale_target.setter
    def scale_target(self, scale_target):
        """Sets the scale_target of this V1alpha1InferenceGraphSpec.

        ScaleTarget specifies the integer target value of the metric type the Autoscaler watches for. concurrency and rps targets are supported by Knative Pod Autoscaler (https://knative.dev/docs/serving/autoscaling/autoscaling-targets/).  # noqa: E501

        :param scale_target: The scale_target of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: int
        """

        self._scale_target = scale_target

    @property
    def service_account_name(self):
        """Gets the service_account_name of this V1alpha1InferenceGraphSpec.  # noqa: E501

        ServiceAccountName specifies the service account name for the InferenceGraph. https://kubernetes.io/docs/tasks/configure-pod-container/configure-service-account/  # noqa: E501

        :return: The service_account_name of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: str
        """
        return self._service_account_name

    @service_account_name.setter
    def service_account_name(self, service_account_name):
        """Sets the service_account_name of this V1alpha1InferenceGraphSpec.

        ServiceAccountName specifies the service account name for the InferenceGraph. https://kubernetes.io/docs/tasks/configure-pod-container/configure-service-account/  # noqa: E501

        :param service_account_name: The service_account_name of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: str
        """

        self._service_account_name = service_account_name

    @property
    def timeout(self):
        """Gets the timeout of this V1alpha1InferenceGraphSpec.  # noqa: E501

        TimeoutSeconds specifies the number of seconds to wait before timing out a request to the component.  # noqa: E501

        :return: The timeout of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: int
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Sets the timeout of this V1alpha1InferenceGraphSpec.

        TimeoutSeconds specifies the number of seconds to wait before timing out a request to the component.  # noqa: E501

        :param timeout: The timeout of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: int
        """

        self._timeout = timeout

    @property
    def tolerations(self):
        """Gets the tolerations of this V1alpha1InferenceGraphSpec.  # noqa: E501

        Toleration specifies the toleration for the InferenceGraph. https://kubernetes.io/docs/concepts/scheduling-eviction/taint-and-toleration/  # noqa: E501

        :return: The tolerations of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :rtype: list[V1Toleration]
        """
        return self._tolerations

    @tolerations.setter
    def tolerations(self, tolerations):
        """Sets the tolerations of this V1alpha1InferenceGraphSpec.

        Toleration specifies the toleration for the InferenceGraph. https://kubernetes.io/docs/concepts/scheduling-eviction/taint-and-toleration/  # noqa: E501

        :param tolerations: The tolerations of this V1alpha1InferenceGraphSpec.  # noqa: E501
        :type: list[V1Toleration]
        """

        self._tolerations = tolerations

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1InferenceGraphSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1InferenceGraphSpec):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1alpha1_serving_runtime_spec import (
    V1alpha1ServingRuntimeSpec,
)  # noqa: E501
from kserve.rest import ApiException


class TestV1alpha1ServingRuntimeSpec(unittest.TestCase):
    """V1alpha1ServingRuntimeSpec unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1alpha1ServingRuntimeSpec
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1alpha1_serving_runtime_spec.V1alpha1ServingRuntimeSpec()  # noqa: E501
        if include_optional:
            return V1alpha1ServingRuntimeSpec(
                affinity=None,
                built_in_adapter=kserve.models.v1alpha1_built_in_adapter.V1alpha1BuiltInAdapter(
                    mem_buffer_bytes=56,
                    model_loading_timeout_millis=56,
                    runtime_management_port=56,
                    server_type="0",
                ),
                containers=[
                    kserve.models.v1alpha1_container.V1alpha1Container(
                        args=["0"],
                        command=["0"],
                        env=[None],
                        image="0",
                        image_pull_policy="0",
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        working_dir="0",
                    )
                ],
                disabled=True,
                grpc_data_endpoint="0",
                grpc_endpoint="0",
                http_data_endpoint="0",
                node_selector={"key": "0"},
                replicas=56,
                storage_helper=kserve.models.v1alpha1_storage_helper.V1alpha1StorageHelper(
                    disabled=True,
                ),
                supported_model_formats=[
                    kserve.models.v1alpha1_supported_model_format.V1alpha1SupportedModelFormat(
                        auto_select=True,
                        name="0",
                        version="0",
                    )
                ],
                tolerations=[None],
            )
        else:
            return V1alpha1ServingRuntimeSpec(
                containers=[
                    kserve.models.v1alpha1_container.V1alpha1Container(
                        args=["0"],
                        command=["0"],
                        env=[None],
                        image="0",
                        image_pull_policy="0",
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        working_dir="0",
                    )
                ],
            )

    def testV1alpha1ServingRuntimeSpec(self):
        """Test V1alpha1ServingRuntimeSpec"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from typing import AsyncGenerator

from fastapi import APIRouter, FastAPI, Request, Response
from fastapi.responses import ORJSONResponse
from fastapi.exceptions import RequestValidationError
from pydantic import TypeAdapter, ValidationError
from starlette.responses import StreamingResponse
from vllm.entrypoints.utils import with_cancellation

from kserve.protocol.rest.openai.types import (
    ChatCompletionRequest,
    CompletionRequest,
    EmbeddingRequest,
    ErrorResponse,
    Model,
    ModelList,
    RerankRequest,
)

from ....errors import ModelNotReady
from .dataplane import OpenAIDataPlane
from .errors import OpenAIError, openai_error_handler

OPENAI_ROUTE_PREFIX = os.environ.get("KSERVE_OPENAI_ROUTE_PREFIX", "/openai")

if len(OPENAI_ROUTE_PREFIX) > 0 and not OPENAI_ROUTE_PREFIX.startswith("/"):
    OPENAI_ROUTE_PREFIX = f"/{OPENAI_ROUTE_PREFIX}"


CreateCompletionRequestAdapter = TypeAdapter(CompletionRequest)
ChatCompletionRequestAdapter = TypeAdapter(ChatCompletionRequest)
EmbeddingRequestAdapter = TypeAdapter(EmbeddingRequest)
RerankRequestAdapter = TypeAdapter(RerankRequest)


class OpenAIEndpoints:
    def __init__(self, dataplane: OpenAIDataPlane):
        self.dataplane = dataplane
        self.start_time = int(time.time())

    @with_cancellation
    async def create_completion(
        self,
        request_body: CompletionRequest,
        raw_request: Request,
        response: Response,
    ) -> Response:
        """Create completion handler.

        Args:
            request_body (CompletionCreateParams): Completion params body.
            raw_request (Request): fastapi request object,
            response (Response): fastapi response object

        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = CreateCompletionRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError from e
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        completion = await self.dataplane.create_completion(
            model_name=model_name,
            request=params,
            raw_request=raw_request,
            headers=raw_request.headers,
            response=response,
        )
        if isinstance(completion, ErrorResponse):
            return ORJSONResponse(
                content=completion.model_dump(), status_code=int(completion.error.code)
            )
        elif isinstance(completion, AsyncGenerator):
            return StreamingResponse(completion, media_type="text/event-stream")
        else:
            return completion

    @with_cancellation
    async def create_chat_completion(
        self,
        request_body: ChatCompletionRequest,
        raw_request: Request,
        response: Response,
    ) -> Response:
        """Create chat completion handler.

        Args:
            request_body (ChatCompletionRequestAdapter): Chat completion params body.
            raw_request (Request): fastapi request object,
            response (Response): fastapi response object

        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = ChatCompletionRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError from e
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        request_headers = raw_request.headers
        completion = await self.dataplane.create_chat_completion(
            model_name=model_name,
            request=request_body,
            raw_request=raw_request,
            headers=request_headers,
            response=response,
        )
        if isinstance(completion, ErrorResponse):
            return ORJSONResponse(
                content=completion.model_dump(), status_code=int(completion.error.code)
            )
        elif isinstance(completion, AsyncGenerator):
            return StreamingResponse(completion, media_type="text/event-stream")
        else:
            return completion

    @with_cancellation
    async def create_embedding(
        self,
        request_body: EmbeddingRequest,
        raw_request: Request,
        response: Response,
    ) -> Response:
        """Create embedding handler.
        Args:
            request_body (EmbeddingRequestAdapter): Embedding params body.
            raw_request (Request): fastapi request object,
            model_name (str): Model name.
        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = EmbeddingRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError from e
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        embedding = await self.dataplane.create_embedding(
            model_name=model_name,
            request=params,
            raw_request=raw_request,
            headers=raw_request.headers,
            response=response,
        )
        if isinstance(embedding, ErrorResponse):
            return ORJSONResponse(
                content=embedding.model_dump(), status_code=int(embedding.error.code)
            )
        elif isinstance(embedding, AsyncGenerator):
            return StreamingResponse(embedding, media_type="text/event-stream")
        else:
            return embedding

    async def create_rerank(
        self,
        raw_request: Request,
        request_body: RerankRequest,
        response: Response,
    ) -> Response:
        """Create rerank handler.
        Args:
            raw_request (Request): fastapi request object,
            model_name (str): Model name.
            request_body (RerankRequestAdapter): Rerank params body.
        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = RerankRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError(errors=e.errors())
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        rerank = await self.dataplane.create_rerank(
            model_name=model_name,
            request=params,
            raw_request=raw_request,
            headers=raw_request.headers,
            response=response,
        )
        if isinstance(rerank, ErrorResponse):
            return ORJSONResponse(
                content=rerank.model_dump(), status_code=int(rerank.error.code)
            )
        elif isinstance(rerank, AsyncGenerator):
            return StreamingResponse(rerank, media_type="text/event-stream")
        else:
            return rerank

    async def models(
        self,
    ) -> ModelList:
        """Create chat completion handler.

        Args:
            raw_request (Request): fastapi request object,

        Returns:
            ModelList: Model response object.
        """
        models = await self.dataplane.models()
        return ModelList(
            object="list",
            data=[
                Model(
                    object="model", id=model.name, created=self.start_time, owned_by=""
                )
                for model in models
            ],
        )

    async def health(self, model_name: str):
        try:
            model_ready = await self.dataplane.model_ready(model_name)
        except Exception as e:
            raise ModelNotReady(model_name) from e
        if not model_ready:
            raise ModelNotReady(model_name)


def register_openai_endpoints(app: FastAPI, dataplane: OpenAIDataPlane):
    endpoints = OpenAIEndpoints(dataplane)
    openai_router = APIRouter(prefix=OPENAI_ROUTE_PREFIX, tags=["OpenAI"])
    openai_router.add_api_route(
        r"/v1/completions",
        endpoints.create_completion,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/chat/completions",
        endpoints.create_chat_completion,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/embeddings",
        endpoints.create_embedding,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/rerank",
        endpoints.create_rerank,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/models",
        endpoints.models,
        methods=["GET"],
    )
    openai_router.add_api_route(
        r"/v1/models/{model_name}", endpoints.health, methods=["GET"]
    )
    app.include_router(openai_router)
    app.add_exception_handler(OpenAIError, openai_error_handler)



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1beta1ModelStorageSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'key': 'str',
        'parameters': 'dict(str, str)',
        'path': 'str',
        'schema_path': 'str'
    }

    attribute_map = {
        'key': 'key',
        'parameters': 'parameters',
        'path': 'path',
        'schema_path': 'schemaPath'
    }

    def __init__(self, key=None, parameters=None, path=None, schema_path=None, local_vars_configuration=None):  # noqa: E501
        """V1beta1ModelStorageSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._key = None
        self._parameters = None
        self._path = None
        self._schema_path = None
        self.discriminator = None

        if key is not None:
            self.key = key
        if parameters is not None:
            self.parameters = parameters
        if path is not None:
            self.path = path
        if schema_path is not None:
            self.schema_path = schema_path

    @property
    def key(self):
        """Gets the key of this V1beta1ModelStorageSpec.  # noqa: E501

        The Storage Key in the secret for this object.  # noqa: E501

        :return: The key of this V1beta1ModelStorageSpec.  # noqa: E501
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key):
        """Sets the key of this V1beta1ModelStorageSpec.

        The Storage Key in the secret for this object.  # noqa: E501

        :param key: The key of this V1beta1ModelStorageSpec.  # noqa: E501
        :type: str
        """

        self._key = key

    @property
    def parameters(self):
        """Gets the parameters of this V1beta1ModelStorageSpec.  # noqa: E501

        Parameters to override the default storage credentials and config.  # noqa: E501

        :return: The parameters of this V1beta1ModelStorageSpec.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """Sets the parameters of this V1beta1ModelStorageSpec.

        Parameters to override the default storage credentials and config.  # noqa: E501

        :param parameters: The parameters of this V1beta1ModelStorageSpec.  # noqa: E501
        :type: dict(str, str)
        """

        self._parameters = parameters

    @property
    def path(self):
        """Gets the path of this V1beta1ModelStorageSpec.  # noqa: E501

        The path to the object in the storage. Note that this path is relative to the storage URI.  # noqa: E501

        :return: The path of this V1beta1ModelStorageSpec.  # noqa: E501
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path of this V1beta1ModelStorageSpec.

        The path to the object in the storage. Note that this path is relative to the storage URI.  # noqa: E501

        :param path: The path of this V1beta1ModelStorageSpec.  # noqa: E501
        :type: str
        """

        self._path = path

    @property
    def schema_path(self):
        """Gets the schema_path of this V1beta1ModelStorageSpec.  # noqa: E501

        The path to the model schema file in the storage.  # noqa: E501

        :return: The schema_path of this V1beta1ModelStorageSpec.  # noqa: E501
        :rtype: str
        """
        return self._schema_path

    @schema_path.setter
    def schema_path(self, schema_path):
        """Sets the schema_path of this V1beta1ModelStorageSpec.

        The path to the model schema file in the storage.  # noqa: E501

        :param schema_path: The schema_path of this V1beta1ModelStorageSpec.  # noqa: E501
        :type: str
        """

        self._schema_path = schema_path

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1beta1ModelStorageSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1beta1ModelStorageSpec):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import json

import numpy as np
import pytest
from orjson import orjson

from kserve import InferRequest, InferInput, InferResponse, InferOutput
from kserve.errors import InvalidInput
from kserve.protocol.grpc.grpc_predict_v2_pb2 import (
    ModelInferRequest,
    InferParameter,
    ModelInferResponse,
)
from kserve.protocol.infer_type import (
    serialize_byte_tensor,
    _contains_fp16_datatype,
    RequestedOutput,
)
from kserve.protocol.rest.v2_datamodels import (
    InferenceRequest,
    RequestInput,
    RequestOutput,
)


class TestInferRequest:
    def test_to_grpc(self):
        infer_req = InferRequest(
            model_name="TestModel",
            model_version="v1",
            request_id="123",
            parameters={"test-str": "dummy", "test-bool": True, "test-int": 100},
            infer_inputs=[
                InferInput(
                    name="input-0",
                    datatype="INT32",
                    shape=[1, 2],
                    data=[1, 2],
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                )
            ],
            request_outputs=[
                RequestedOutput(
                    name="output-0",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
                RequestedOutput(
                    name="output-1",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
            ],
        )
        expected = ModelInferRequest(
            model_name="TestModel",
            model_version="v1",
            id="123",
            parameters={
                "test-str": InferParameter(string_param="dummy"),
                "test-bool": InferParameter(bool_param=True),
                "test-int": InferParameter(int64_param=100),
            },
            inputs=[
                {
                    "name": "input-0",
                    "shape": [1, 2],
                    "datatype": "INT32",
                    "contents": {"int_contents": [1, 2]},
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                }
            ],
            outputs=[
                {
                    "name": "output-0",
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                },
                {
                    "name": "output-1",
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                },
            ],
        )
        res = infer_req.to_grpc()
        assert res == expected

    def test_from_grpc(self):
        infer_req = ModelInferRequest(
            model_name="TestModel",
            id="123",
            parameters={
                "test-str": InferParameter(string_param="dummy"),
                "test-bool": InferParameter(bool_param=True),
                "test-int": InferParameter(int64_param=100),
            },
            inputs=[
                {
                    "name": "input-0",
                    "shape": [1, 2],
                    "datatype": "INT32",
                    "contents": {"int_contents": [1, 2]},
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                }
            ],
            outputs=[
                {
                    "name": "output-0",
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                },
                {
                    "name": "output-1",
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                },
            ],
        )
        expected = InferRequest(
            model_name="TestModel",
            request_id="123",
            parameters={
                "test-str": InferParameter(string_param="dummy"),
                "test-bool": InferParameter(bool_param=True),
                "test-int": InferParameter(int64_param=100),
            },
            infer_inputs=[
                InferInput(
                    name="input-0",
                    datatype="INT32",
                    shape=[1, 2],
                    data=[1, 2],
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                )
            ],
            request_outputs=[
                RequestedOutput(
                    name="output-0",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
                RequestedOutput(
                    name="output-1",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
            ],
            from_grpc=True,
        )
        res = InferRequest.from_grpc(infer_req)
        assert res == expected

    def test_to_rest_with_mixed_binary_data(self):
        infer_input_1 = InferInput(
            name="input1",
            shape=[3],
            datatype="INT32",
            data=np.array([1, 2, 3], dtype=np.int32),
            parameters={
                "test-str": "dummy",
            },
        )
        infer_input_2 = InferInput(
            name="input2",
            shape=[1],
            datatype="BYTES",
            data=None,
            parameters={
                "test-int": 2,
            },
        )
        infer_input_2.set_data_from_numpy(
            np.array(["test"], dtype=np.object_), binary_data=True
        )
        infer_input_3 = InferInput(
            name="input3",
            shape=[3],
            datatype="FP16",
            data=None,
        )
        infer_input_3.set_data_from_numpy(
            np.array([1.2, 2.2, 3.2], dtype=np.float16), binary_data=True
        )
        infer_request = InferRequest(
            request_id="4be4e82f-5500-420a-a5c5-ac86841e271b",
            model_name="test_model",
            infer_inputs=[infer_input_1, infer_input_2, infer_input_3],
            request_outputs=[
                RequestedOutput(
                    name="output-0",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
                RequestedOutput(
                    name="output-1",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
            ],
        )
        infer_request_bytes, json_length = infer_request.to_rest()
        assert isinstance(infer_request_bytes, bytes)
        assert (
            infer_request_bytes
            == b'{"id":"4be4e82f-5500-420a-a5c5-ac86841e271b","model_name":"test_model","inputs":[{"name":"input1","shape":[3],"datatype":"INT32","parameters":{"test-str":"dummy"},"data":[1,2,3]},{"name":"input2","shape":[1],"datatype":"BYTES","parameters":{"test-int":2,"binary_data_size":8}},{"name":"input3","shape":[3],"datatype":"FP16","parameters":{"binary_data_size":6}}],"outputs":[{"name":"output-0","parameters":{"test-str":"dummy","test-bool":true,"test-int":100}},{"name":"output-1","parameters":{"test-str":"dummy","test-bool":true,"test-int":100}}]}\x04\x00\x00\x00test\xcd<f@fB'
        )
        assert json_length == 546

    def test_to_rest_with_fp16_not_as_binary_data(self):
        infer_input_1 = InferInput(
            name="input1",
            shape=[3],
            datatype="INT32",
            data=None,
        )
        infer_input_1.set_data_from_numpy(
            np.array([1, 2, 3], dtype=np.int32), binary_data=False
        )
        infer_input_2 = InferInput(
            name="input2",
            shape=[1],
            datatype="BYTES",
            data=None,
        )
        infer_input_2.set_data_from_numpy(
            np.array(["test"], dtype=np.object_), binary_data=True
        )
        infer_input_3 = InferInput(
            name="input3",
            shape=[3],
            datatype="FP16",
            data=None,
        )
        infer_input_3.set_data_from_numpy(
            np.array([1.2, 2.2, 3.2], dtype=np.float16), binary_data=False
        )
        infer_request = InferRequest(
            request_id="4be4e82f-5500-420a-a5c5-ac86841e271b",
            model_name="test_model",
            infer_inputs=[infer_input_1, infer_input_2, infer_input_3],
        )
        with pytest.raises(InvalidInput):
            infer_request_bytes, json_length = infer_request.to_rest()

    def test_to_rest_with_no_binary_data(self):
        infer_input_1 = InferInput(
            name="input1",
            shape=[3],
            datatype="INT32",
            data=[1, 2, 3],
        )
        infer_input_2 = InferInput(
            name="input2",
            shape=[1],
            datatype="BYTES",
            data=["test"],
        )
        infer_input_3 = InferInput(
            name="input3",
            shape=[3],
            datatype="FP32",
            data=[1.2, 2.2, 3.2],
            parameters={
                "test-int": 2,
            },
        )
        infer_request = InferRequest(
            request_id="4be4e82f-5500-420a-a5c5-ac86841e271b",
            model_name="test_model",
            infer_inputs=[infer_input_1, infer_input_2, infer_input_3],
            request_outputs=[
                RequestedOutput(
                    name="output-0",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
                RequestedOutput(
                    name="output-1",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
            ],
        )
        infer_request_dict, json_length = infer_request.to_rest()
        assert isinstance(infer_request_dict, dict)
        assert infer_request_dict == {
            "id": "4be4e82f-5500-420a-a5c5-ac86841e271b",
            "model_name": "test_model",
            "inputs": [
                {
                    "name": "input1",
                    "shape": [3],
                    "datatype": "INT32",
                    "data": [1, 2, 3],
                },
                {"name": "input2", "shape": [1], "datatype": "BYTES", "data": ["test"]},
                {
                    "name": "input3",
                    "shape": [3],
                    "datatype": "FP32",
                    "parameters": {"test-int": 2},
                    "data": [1.2, 2.2, 3.2],
                },
            ],
            "outputs": [
                {
                    "name": "output-0",
                    "parameters": {
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                },
                {
                    "name": "output-1",
                    "parameters": {
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                },
            ],
        }
        assert json_length is None

    def test_to_rest_with_parameters(self):
        infer_input = InferInput(
            name="input1",
            shape=[3],
            datatype="INT32",
            data=None,
            parameters={"param1": "value1"},
        )
        infer_input.set_data_from_numpy(
            np.array([1, 2, 3], dtype=np.int32), binary_data=True
        )
        infer_request = InferRequest(
            model_name="test_model",
            infer_inputs=[infer_input],
            parameters={"request_param1": "request_value1"},
        )
        infer_request_bytes, json_length = infer_request.to_rest()
        assert infer_request_bytes is not None
        assert json_length > 0
        infer_request_dict = orjson.loads(infer_request_bytes[:json_length])
        assert infer_request_dict["parameters"]["request_param1"] == "request_value1"
        assert infer_request_dict["inputs"][0]["parameters"]["param1"] == "value1"

    def test_infer_request_to_rest_missing_data_field(self):
        infer_input = InferInput(
            name="input1",
            shape=[1],
            datatype="INT32",
            parameters={"binary_data_size": 12},
        )
        infer_request = InferRequest(
            model_name="test_model",
            infer_inputs=[infer_input],
        )
        with pytest.raises(InvalidInput):
            infer_request_bytes, json_length = infer_request.to_rest()

    def test_infer_request_from_bytes_valid_input(self):
        infer_input_1 = InferInput(
            name="input1",
            shape=[3],
            datatype="INT32",
            data=[1, 2, 3],
        )
        infer_input_2 = InferInput(
            name="input2",
            shape=[3],
            datatype="FP16",
            data=None,
            parameters={},
        )
        infer_input_2.set_data_from_numpy(
            np.array([1.1, 2.9, 3.4], dtype=np.float16), binary_data=True
        )
        infer_request = InferRequest(
            request_id="abc",
            model_name="test_model",
            infer_inputs=[infer_input_1, infer_input_2],
        )
        expected = copy.deepcopy(infer_request)
        infer_request_bytes, json_length = infer_request.to_rest()
        infer_request_from_bytes = InferRequest.from_bytes(
            infer_request_bytes, json_length, "test_model"
        )
        infer_request_from_bytes.inputs[1].set_data_from_numpy(
            infer_request_from_bytes.inputs[1].as_numpy(), binary_data=True
        )
        assert expected == infer_request_from_bytes

    def test_infer_request_from_bytes_with_requested_outputs(self):
        infer_input_1 = InferInput(
            name="input1",
            shape=[3],
            datatype="INT32",
            data=[1, 2, 3],
        )
        infer_input_2 = InferInput(
            name="input2",
            shape=[3],
            datatype="FP16",
            data=None,
            parameters={},
        )
        infer_input_2.set_data_from_numpy(
            np.array([1.1, 2.9, 3.4], dtype=np.float16), binary_data=True
        )
        infer_request = InferRequest(
            request_id="abc",
            model_name="test_model",
            infer_inputs=[infer_input_1, infer_input_2],
            request_outputs=[
                RequestedOutput(
                    name="output-0",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
                RequestedOutput(
                    name="output-1",
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                ),
            ],
        )
        expected = copy.deepcopy(infer_request)
        infer_request_bytes, json_length = infer_request.to_rest()
        infer_request_from_bytes = InferRequest.from_bytes(
            infer_request_bytes, json_length, "test_model"
        )
        infer_request_from_bytes.inputs[1].set_data_from_numpy(
            infer_request_from_bytes.inputs[1].as_numpy(), binary_data=True
        )
        assert expected == infer_request_from_bytes

    def test_infer_request_from_bytes_invalid_json(self):
        with pytest.raises(InvalidInput):
            InferRequest.from_bytes(
                b'{"id": "1", "inputs": [{"name": "input1", "shape": [1], "datatype": "INT32", "data": [1]}',
                100,
                "test_model",
            )

    def test_infer_request_from_bytes_missing_data_field(self):
        infer_request_bytes = b'{"id": "1", "inputs": [{"name": "input1", "shape": [1], "datatype": "INT32"'
        with pytest.raises(InvalidInput):
            InferRequest.from_bytes(
                infer_request_bytes, len(infer_request_bytes), "test_model"
            )

    def test_infer_request_from_bytes_missing_binary_data_size(self):
        infer_request_bytes = b'{"id":"509c5da9-80d4-46e8-a50c-0bba2b9d76f8","inputs":[{"name":"input1","shape":[3],"datatype":"INT32"}]}\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00'

        with pytest.raises(InvalidInput):
            InferRequest.from_bytes(
                infer_request_bytes,
                len(
                    b'{"id":"509c5da9-80d4-46e8-a50c-0bba2b9d76f8","inputs":[{"name":"input1","shape":[3],"datatype":"INT32"}]}'
                ),
                "test_model",
            )

    def test_infer_request_from_bytes_fp16_data_via_json(self):

        infer_request_bytes = b'{"id": "1", "inputs": [{"name": "input1", "shape": [1], "datatype": "FP16", "data": [1]}'
        with pytest.raises(InvalidInput):
            InferRequest.from_bytes(
                infer_request_bytes, len(infer_request_bytes), "test_model"
            )

    def test_from_inference_request_with_valid_input(self):
        inference_request = InferenceRequest(
            id="test_id",
            inputs=[
                RequestInput(
                    name="input1",
                    shape=[3],
                    datatype="INT32",
                    data=[1, 2, 3],
                    parameters={
                        "test-str": "dummy",
                    },
                ),
                RequestInput(
                    name="input2",
                    shape=[3],
                    datatype="INT32",
                    data=[1, 2, 3],
                    parameters={
                        "test-int": 1,
                    },
                ),
            ],
            outputs=[
                RequestOutput(name="output1", parameters={"binary_data": True}),
                RequestOutput(name="output2", parameters={"binary_data": False}),
            ],
            parameters={"test-bool": True},
        )
        infer_request = InferRequest.from_inference_request(
            inference_request, "test_model"
        )
        assert infer_request.model_name == "test_model"
        assert infer_request.id == "test_id"
        assert infer_request.parameters == {"test-bool": True}
        assert len(infer_request.inputs) == 2
        assert infer_request.inputs[0].name == "input1"
        assert infer_request.inputs[0].shape == [3]
        assert infer_request.inputs[0].datatype == "INT32"
        assert infer_request.inputs[0].data == [1, 2, 3]
        assert infer_request.inputs[0].parameters == {"test-str": "dummy"}
        assert len(infer_request.request_outputs) == 2
        assert infer_request.request_outputs[0].name == "output1"
        assert infer_request.request_outputs[0].parameters == {"binary_data": True}
        assert infer_request.request_outputs[1].name == "output2"
        assert infer_request.request_outputs[1].parameters == {"binary_data": False}

    def test_from_inference_request_with_fp16_data(self):
        inference_request = InferenceRequest(
            id="test_id",
            inputs=[
                RequestInput(
                    name="input1", shape=[3], datatype="FP16", data=[1.0, 2.0, 3.0]
                )
            ],
        )
        with pytest.raises(InvalidInput):
            InferRequest.from_inference_request(inference_request, "test_model")

    def test_from_inference_request_with_no_requested_outputs(self):
        inference_request = InferenceRequest(
            id="test_id",
            inputs=[
                RequestInput(name="input1", shape=[3], datatype="INT32", data=[1, 2, 3])
            ],
        )
        infer_request = InferRequest.from_inference_request(
            inference_request, "test_model"
        )
        assert infer_request.model_name == "test_model"
        assert infer_request.id == "test_id"
        assert len(infer_request.inputs) == 1
        assert infer_request.inputs[0].name == "input1"
        assert infer_request.inputs[0].shape == [3]
        assert infer_request.inputs[0].datatype == "INT32"
        assert infer_request.inputs[0].data == [1, 2, 3]
        assert infer_request.request_outputs is None


class TestInferResponse:
    def test_to_grpc(self):
        infer_res = InferResponse(
            model_name="TestModel",
            response_id="123",
            model_version="v1",
            parameters={"test-str": "dummy", "test-bool": True, "test-int": 100},
            infer_outputs=[
                InferOutput(
                    name="output-0",
                    datatype="INT32",
                    shape=[1, 2],
                    data=[1, 2],
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                    },
                )
            ],
        )
        expected = ModelInferResponse(
            model_name="TestModel",
            id="123",
            model_version="v1",
            parameters={
                "test-str": InferParameter(string_param="dummy"),
                "test-bool": InferParameter(bool_param=True),
                "test-int": InferParameter(int64_param=100),
            },
            outputs=[
                {
                    "name": "output-0",
                    "shape": [1, 2],
                    "datatype": "INT32",
                    "contents": {"int_contents": [1, 2]},
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                }
            ],
        )
        res = infer_res.to_grpc()
        assert res == expected

    def test_from_grpc(self):
        infer_res = ModelInferResponse(
            model_name="TestModel",
            id="123",
            model_version="v1",
            parameters={
                "test-str": InferParameter(string_param="dummy"),
                "test-bool": InferParameter(bool_param=True),
                "test-int": InferParameter(int64_param=100),
            },
            outputs=[
                {
                    "name": "output-0",
                    "shape": [1, 2],
                    "datatype": "INT32",
                    "contents": {"int_contents": [1, 2]},
                    "parameters": {
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                }
            ],
        )
        expected = InferResponse(
            model_name="TestModel",
            response_id="123",
            model_version="v1",
            parameters={
                "test-str": InferParameter(string_param="dummy"),
                "test-bool": InferParameter(bool_param=True),
                "test-int": InferParameter(int64_param=100),
            },
            infer_outputs=[
                InferOutput(
                    name="output-0",
                    datatype="INT32",
                    shape=[1, 2],
                    data=[1, 2],
                    parameters={
                        "test-str": InferParameter(string_param="dummy"),
                        "test-bool": InferParameter(bool_param=True),
                        "test-int": InferParameter(int64_param=100),
                    },
                )
            ],
            from_grpc=True,
        )
        res = InferResponse.from_grpc(infer_res)
        assert res == expected

    def test_infer_response_to_rest_with_binary_data(self):
        data = np.array(
            [[1.2, 2.2, 3.2, 4.1], [1.5, 2.6, 3.787, 4.54]], dtype=np.float16
        )
        fp16_output = InferOutput(
            name="fp16_output",
            shape=list(data.shape),
            datatype="FP16",
            data=None,
            parameters=None,
        )
        fp32_output = InferOutput(
            name="fp32_output",
            shape=list(data.shape),
            datatype="FP32",
            data=None,
            parameters=None,
        )
        fp16_output.set_data_from_numpy(data, binary_data=True)
        fp32_output.set_data_from_numpy(data.astype(np.float32), binary_data=True)
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[fp16_output, fp32_output],
            requested_outputs=[
                RequestedOutput(name="fp16_output", parameters={"binary_data": True}),
                RequestedOutput(name="fp32_output", parameters={"binary_data": True}),
            ],
        )
        result, json_length = infer_response.to_rest()
        assert isinstance(result, bytes)
        assert (
            result
            == b'{"id":"1","model_name":"test_model","model_version":null,"outputs":[{"name":"fp16_output","shape":[2,4],"datatype":"FP16","parameters":{"binary_data_size":16}},{"name":"fp32_output","shape":[2,4],"datatype":"FP32","parameters":{"binary_data_size":32}}]}\xcd<f@fB\x1aD\x00>3A\x93C\x8aD\x00\xa0\x99?\x00\xc0\x0c@\x00\xc0L@\x00@\x83@\x00\x00\xc0?\x00`&@\x00`r@\x00@\x91@'
        )
        assert json_length == 253

    def test_infer_response_to_rest_without_binary_data(self):
        int32_output = InferOutput(
            name="int32_output",
            shape=[1],
            datatype="INT32",
            data=[1],
            parameters=None,
        )
        str_output = InferOutput(
            name="str_output",
            shape=[1],
            datatype="BYTES",
            data=["test"],
            parameters=None,
        )
        uint_data = np.array([[1, 2]], dtype=np.uint32)
        uint_output = InferOutput(
            name="uint_output",
            shape=list(uint_data.shape),
            datatype="UINT32",
            parameters=None,
            data=uint_data,
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[int32_output, str_output, uint_output],
        )

        result, json_length = infer_response.to_rest()
        assert isinstance(result, dict)
        assert result == {
            "id": "1",
            "model_name": "test_model",
            "model_version": None,
            "outputs": [
                {
                    "name": "int32_output",
                    "shape": [1],
                    "datatype": "INT32",
                    "data": [1],
                },
                {
                    "name": "str_output",
                    "shape": [1],
                    "datatype": "BYTES",
                    "data": ["test"],
                },
                {
                    "name": "uint_output",
                    "shape": [1, 2],
                    "datatype": "UINT32",
                    "data": [1, 2],
                },
            ],
        }
        assert json_length is None

    def test_infer_response_to_rest_with_mixed_binary_data(self):
        infer_output1 = InferOutput(
            name="output1",
            shape=[1],
            datatype="FP16",
            data=np.array([1], dtype=np.float16),
            parameters=None,
        )
        infer_output1.set_data_from_numpy(
            np.array([1], dtype=np.float16), binary_data=True
        )
        infer_output2 = InferOutput(
            name="output2",
            shape=[1],
            datatype="INT32",
            data=[1],
            parameters=None,
        )
        infer_output2.set_data_from_numpy(
            np.array([1], dtype=np.int32), binary_data=False
        )
        infer_output3 = InferOutput(
            name="output3",
            shape=[1],
            datatype="BYTES",
            data=None,
            parameters=None,
        )
        infer_output3.set_data_from_numpy(
            np.array(["test"], dtype=np.object_), binary_data=True
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[infer_output1, infer_output2, infer_output3],
            requested_outputs=[
                RequestedOutput(name="output1", parameters={"binary_data": True}),
                RequestedOutput(name="output2", parameters={"binary_data": True}),
                RequestedOutput(name="output3", parameters={"binary_data": False}),
            ],
        )
        result, json_length = infer_response.to_rest()
        assert isinstance(result, bytes)
        assert (
            result
            == b'{"id":"1","model_name":"test_model","model_version":null,"outputs":[{"name":"output1","shape":[1],"datatype":"FP16","parameters":{"binary_data_size":2}},{"name":"output2","shape":[1],"datatype":"INT32","parameters":{"binary_data_size":4}},{"name":"output3","shape":[1],"datatype":"BYTES","data":["test"]}]}\x00<\x01\x00\x00\x00'
        )
        assert json_length == 306

    def test_infer_response_to_rest_with_binary_output_data_true(self):
        infer_output1 = InferOutput(
            name="output1",
            shape=[1],
            datatype="FP16",
            data=[1],
            parameters=None,
        )
        infer_output2 = InferOutput(
            name="output2",
            shape=[1],
            datatype="FP16",
            data=None,
            parameters=None,
        )
        infer_output2.set_data_from_numpy(
            np.array([1], dtype=np.float16), binary_data=True
        )
        infer_output3 = InferOutput(
            name="output3",
            shape=[1],
            datatype="BYTES",
            data=[b"test"],
            parameters=None,
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[infer_output1, infer_output2, infer_output3],
            use_binary_outputs=True,
        )
        result, json_length = infer_response.to_rest()
        assert isinstance(result, bytes)
        assert (
            result
            == b'{"id":"1","model_name":"test_model","model_version":null,"outputs":[{"name":"output1","shape":[1],"datatype":"FP16","parameters":{"binary_data_size":2}},{"name":"output2","shape":[1],"datatype":"FP16","parameters":{"binary_data_size":2}},{"name":"output3","shape":[1],"datatype":"BYTES","parameters":{"binary_data_size":8}}]}\x00<\x00<\x04\x00\x00\x00test'
        )
        assert json_length == 325

    def test_infer_response_to_rest_with_binary_output_data_precedence(
        self,
    ):
        infer_output1 = InferOutput(
            name="output1",
            shape=[1],
            datatype="FP16",
            data=np.array([1], dtype=np.float16),
            parameters=None,
        )
        infer_output1.set_data_from_numpy(
            np.array([1], dtype=np.float16), binary_data=True
        )
        infer_output2 = InferOutput(
            name="output2",
            shape=[1],
            datatype="INT32",
            data=[1],
            parameters=None,
        )
        infer_output3 = InferOutput(
            name="output3",
            shape=[1],
            datatype="BYTES",
            data=["test"],
            parameters=None,
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[infer_output1, infer_output2, infer_output3],
            use_binary_outputs=True,
            requested_outputs=[
                RequestedOutput(name="output1", parameters={"binary_data": True}),
                RequestedOutput(name="output2", parameters={"binary_data": False}),
                RequestedOutput(name="output3", parameters={"binary_data": True}),
            ],
        )
        result, json_length = infer_response.to_rest()
        assert isinstance(result, bytes)
        assert (
            result
            == b'{"id":"1","model_name":"test_model","model_version":null,"outputs":[{"name":"output1","shape":[1],"datatype":"FP16","parameters":{"binary_data_size":2}},{"name":"output2","shape":[1],"datatype":"INT32","data":[1]},{"name":"output3","shape":[1],"datatype":"BYTES","parameters":{"binary_data_size":8}}]}\x00<\x04\x00\x00\x00test'
        )
        assert json_length == 301

    def test_infer_response_to_rest_with_raw_data_with_binary_data_false(
        self,
    ):
        infer_output = InferOutput(
            name="output1",
            shape=[3],
            datatype="INT32",
        )
        raw_data = b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00"
        infer_output._raw_data = raw_data
        infer_request = InferResponse(
            response_id="4be4e82f-5500-420a-a5c5-ac86841e271b",
            model_name="test_model",
            infer_outputs=[infer_output],
        )
        infer_response, json_length = infer_request.to_rest()
        assert infer_response == {
            "id": "4be4e82f-5500-420a-a5c5-ac86841e271b",
            "model_name": "test_model",
            "model_version": None,
            "outputs": [
                {
                    "name": "output1",
                    "shape": [3],
                    "datatype": "INT32",
                    "data": [1, 2, 3],
                }
            ],
        }
        assert json_length is None

    def test_infer_response_to_rest_missing_data_field(self):
        infer_output = InferOutput(
            name="input1",
            shape=[1],
            datatype="INT32",
            parameters={"binary_data_size": 12},
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[infer_output],
        )
        with pytest.raises(InvalidInput):
            _, _ = infer_response.to_rest()

    def test_infer_response_from_bytes_happy_path(self):
        model_name = "test_model"
        response_bytes = b'{"id": "1", "model_name": "test_model", "outputs": [{"name": "output1", "shape": [1], "datatype": "INT32", "data": [1]}]}'
        json_length = len(response_bytes)

        infer_response = InferResponse.from_bytes(
            response_bytes,
            json_length,
        )

        assert infer_response.id == "1"
        assert infer_response.model_name == model_name
        assert len(infer_response.outputs) == 1
        assert infer_response.outputs[0].name == "output1"
        assert infer_response.outputs[0].shape == [1]
        assert infer_response.outputs[0].datatype == "INT32"
        assert infer_response.outputs[0].data == [1]

    def test_infer_response_from_bytes_with_binary_data(self):
        serialized_str_data = serialize_byte_tensor(
            np.array([b"cat", b"dog", b"bird", b"fish"], dtype=np.object_)
        ).item()
        model_name = "test_model"
        response_bytes = json.dumps(
            {
                "model_name": model_name,
                "id": "1",
                "outputs": [
                    {
                        "name": "output1",
                        "shape": [4],
                        "datatype": "BYTES",
                        "parameters": {"binary_data_size": len(serialized_str_data)},
                    }
                ],
            }
        ).encode()
        json_length = len(response_bytes)

        infer_response = InferResponse.from_bytes(
            response_bytes + serialized_str_data,
            json_length,
        )

        assert infer_response.id == "1"
        assert infer_response.model_name == model_name
        assert len(infer_response.outputs) == 1
        assert infer_response.outputs[0].name == "output1"
        assert infer_response.outputs[0].shape == [4]
        assert infer_response.outputs[0].datatype == "BYTES"
        assert infer_response.outputs[0].data == ["cat", "dog", "bird", "fish"]

    def test_infer_response_from_bytes_with_missing_data(self):
        response_bytes = b'{"id": "1", "model_name": "test_model", "outputs": [{"name": "output1", "shape": [1], "datatype": "INT32"}]}'
        json_length = len(response_bytes)

        with pytest.raises(InvalidInput):
            InferResponse.from_bytes(response_bytes, json_length)

    def test_infer_response_get_output_by_name_returns_correct_output(self):
        infer_output1 = InferOutput(
            name="output1", shape=[1], datatype="INT32", data=[1]
        )
        infer_output2 = InferOutput(
            name="output2", shape=[1], datatype="INT32", data=[2]
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[infer_output1, infer_output2],
        )

        result = infer_response.get_output_by_name("output2")

        assert result == infer_output2

    def test_infer_response_get_output_by_name_returns_none_for_non_existent_output(
        self,
    ):
        infer_output1 = InferOutput(
            name="output1", shape=[1], datatype="INT32", data=[1]
        )
        infer_output2 = InferOutput(
            name="output2", shape=[1], datatype="INT32", data=[2]
        )
        infer_response = InferResponse(
            response_id="1",
            model_name="test_model",
            infer_outputs=[infer_output1, infer_output2],
        )

        result = infer_response.get_output_by_name("output3")

        assert result is None


def test_contains_fp16_datatype_with_fp16_output():
    infer_output1 = InferOutput(name="output1", shape=[1], datatype="FP16", data=[1])
    infer_output2 = InferOutput(name="output2", shape=[1], datatype="INT32", data=[2])
    infer_response = InferResponse(
        response_id="1",
        model_name="test_model",
        infer_outputs=[infer_output1, infer_output2],
    )

    assert _contains_fp16_datatype(infer_response) is True


def test_contains_fp16_datatype_without_fp16_output():
    infer_output = InferOutput(name="output1", shape=[1], datatype="INT32", data=[1])
    infer_response = InferResponse(
        response_id="1", model_name="test_model", infer_outputs=[infer_output]
    )

    assert _contains_fp16_datatype(infer_response) is False


def test_contains_fp16_datatype_with_no_outputs():
    infer_response = InferResponse(
        response_id="1", model_name="test_model", infer_outputs=[]
    )

    assert _contains_fp16_datatype(infer_response) is False



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Any, Dict, Optional, cast, AsyncGenerator, Union
from fastapi import Request

from kserve.protocol.rest.openai.types import (
    ChatCompletionRequest,
    CompletionRequest,
    ChatCompletion,
    ChatCompletionChoice,
    ChatCompletionLogProb,
    ChatCompletionLogProbs,
    ChatCompletionChunk,
    ChatMessage,
    ChatCompletionLogProbsContent,
    ChoiceDelta,
    ChunkChoice,
    Completion,
    CompletionChoice,
    CompletionChunk,
    CompletionChunkChoice,
    CompletionLogProbs,
    ErrorResponse,
)

from kserve.errors import InvalidInput
from kserve.protocol.rest.openai.openai_model import AsyncMappingIterator

from .openai_model import (
    OpenAIGenerativeModel,
    ChatPrompt,
)


class OpenAIChatAdapterModel(OpenAIGenerativeModel):
    """
    A helper on top the OpenAI model that automatically maps chat completion requests (/v1/chat/completions)
    to completion requests (/v1/completions).

    Users should extend this model and implement the abstract methods in order to expose these endpoints.
    """

    @abstractmethod
    def apply_chat_template(
        self,
        request: ChatCompletionRequest,
    ) -> ChatPrompt:
        """
        Given a list of chat completion messages, convert them to a prompt.
        """
        pass

    @classmethod
    def chat_completion_params_to_completion_params(
        cls, request: ChatCompletionRequest, prompt: str
    ) -> CompletionRequest:

        return CompletionRequest(
            prompt=prompt,
            model=request.model,
            frequency_penalty=request.frequency_penalty,
            logit_bias=request.logit_bias,
            max_tokens=request.max_tokens,
            n=request.n,
            presence_penalty=request.presence_penalty,
            seed=request.seed,
            stop=request.stop,
            stream=request.stream,
            temperature=request.temperature,
            top_p=request.top_p,
            user=request.user,
            logprobs=request.top_logprobs,
        )

    @classmethod
    def to_choice_logprobs(cls, logprobs: CompletionLogProbs) -> ChatCompletionLogProbs:
        chat_completion_logprobs = []
        for i in range(len(logprobs.tokens)):
            token = logprobs.tokens[i]
            token_logprob = logprobs.token_logprobs[i]
            top_logprobs_dict = logprobs.top_logprobs[i]
            top_logprobs = [
                ChatCompletionLogProb(
                    token=token,
                    bytes=[int(b) for b in token.encode("utf8")],
                    logprob=logprob,
                )
                for token, logprob in top_logprobs_dict.items()
            ]
            chat_completion_logprobs.append(
                ChatCompletionLogProbsContent(
                    token=token,
                    bytes=[int(b) for b in token.encode("utf8")],
                    logprob=token_logprob,
                    top_logprobs=top_logprobs,
                )
            )

        return ChatCompletionLogProbs(content=chat_completion_logprobs)

    @classmethod
    def to_chat_completion_choice(
        cls, completion_choice: CompletionChoice, role: str
    ) -> ChatCompletionChoice:
        # translate Token -> ChatCompletionTokenLogprob
        choice_logprobs = (
            cls.to_choice_logprobs(completion_choice.logprobs)
            if completion_choice.logprobs is not None
            else None
        )
        return ChatCompletionChoice(
            index=0,
            finish_reason=completion_choice.finish_reason,
            logprobs=choice_logprobs,
            message=ChatMessage(content=completion_choice.text, role=role),
        )

    @classmethod
    def to_chat_completion_chunk_choice(
        cls, completion_choice: CompletionChunkChoice, role: str
    ) -> ChunkChoice:
        # translate Token -> ChatCompletionTokenLogprob
        choice_logprobs = (
            cls.to_choice_logprobs(completion_choice.logprobs)
            if completion_choice.logprobs is not None
            else None
        )
        choice_logprobs = (
            ChatCompletionLogProbs(content=choice_logprobs.content)
            if choice_logprobs is not None
            else None
        )
        return ChunkChoice(
            delta=ChoiceDelta(content=completion_choice.text, role=role),
            index=0,
            finish_reason=completion_choice.finish_reason,
            logprobs=choice_logprobs,
        )

    @classmethod
    def completion_to_chat_completion(
        cls, completion: Completion, role: str
    ) -> ChatCompletion:
        completion_choice = (
            completion.choices[0] if len(completion.choices) > 0 else None
        )
        choices = (
            [cls.to_chat_completion_choice(completion_choice, role)]
            if completion_choice is not None
            else []
        )
        return ChatCompletion(
            id=completion.id,
            choices=choices,
            created=completion.created,
            model=completion.model,
            object="chat.completion",
            system_fingerprint=completion.system_fingerprint,
            usage=completion.usage,
        )

    @classmethod
    def completion_to_chat_completion_chunk(
        cls, completion: Completion, role: str
    ) -> ChatCompletionChunk:
        completion_choice = (
            completion.choices[0] if len(completion.choices) > 0 else None
        )
        choices = (
            [cls.to_chat_completion_chunk_choice(completion_choice, role)]
            if completion_choice is not None
            else []
        )
        return ChatCompletionChunk(
            id=completion.id,
            choices=choices,
            created=completion.created,
            model=completion.model,
            object="chat.completion.chunk",
            system_fingerprint=completion.system_fingerprint,
        )

    async def create_chat_completion(
        self,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], ChatCompletion, ErrorResponse]:

        if request.n != 1:
            raise InvalidInput("n != 1 is not supported")

        # Convert the messages into a prompt
        chat_prompt = self.apply_chat_template(request)
        # Translate the chat completion request to a completion request
        completion_params = self.chat_completion_params_to_completion_params(
            request, chat_prompt.prompt
        )

        if not request.stream:
            completion = cast(
                Completion, await self.create_completion(completion_params, raw_request)
            )
            return self.completion_to_chat_completion(
                completion, chat_prompt.response_role
            )
        else:
            completion_iterator = await self.create_completion(
                completion_params, raw_request
            )

            def mapper(completion_str: str) -> ChatCompletionChunk:

                chunk = completion_str.removeprefix("data: ")
                if chunk == "[DONE]\n\n":
                    return

                completion = CompletionChunk.model_validate_json(chunk)

                return self.completion_to_chat_completion_chunk(
                    completion, chat_prompt.response_role
                )

            completion = AsyncMappingIterator(
                iterator=completion_iterator, mapper=mapper
            )

            async def stream_results() -> AsyncGenerator[str, None]:
                async for chunk in completion:
                    yield f"data: {chunk.model_dump_json()}\n\n"
                yield "data: [DONE]\n\n"

            return stream_results()



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from kserve import ModelRepository, Model
from kserve.protocol.rest.openai import (
    CompletionRequest,
    OpenAIGenerativeModel,
    OpenAIEncoderModel,
)
from unittest.mock import patch
from kserve.protocol.rest.openai.types import (
    ChatCompletion,
    ChatCompletionChunk,
    Completion,
    Embedding,
    EmbeddingRequest,
    Rerank,
    RerankRequest,
)
from typing import AsyncIterator, Union


class DummyOpenAIGenerativeModel(OpenAIGenerativeModel):
    async def create_completion(
        self, params: CompletionRequest
    ) -> Union[Completion, AsyncIterator[Completion]]:
        pass

    async def create_chat_completion(
        self, params: CompletionRequest
    ) -> Union[ChatCompletion, AsyncIterator[ChatCompletionChunk]]:
        pass


class DummyOpenAIEncoderModel(OpenAIEncoderModel):
    async def create_embedding(self, params: EmbeddingRequest) -> Embedding:
        pass

    async def create_rerank(self, params: RerankRequest) -> Rerank:
        pass


def test_adding_kserve_model():
    repo = ModelRepository()
    model = Model(name="kserve-model")
    repo.update(model)

    actual = repo.get_model("kserve-model")

    assert actual is not None
    assert isinstance(actual, Model)
    assert actual.name == "kserve-model"

    repo.update(model, name="additional-model-name")
    actual = repo.get_model("additional-model-name")
    assert actual is not None
    assert isinstance(actual, Model)
    assert actual is model


def test_adding_openai_completion_model():
    repo = ModelRepository()
    repo.update(DummyOpenAIGenerativeModel(name="openai-completion-model"))

    actual = repo.get_model("openai-completion-model")

    assert actual is not None
    assert isinstance(actual, OpenAIGenerativeModel)
    assert actual.name == "openai-completion-model"


def test_adding_openai_embedding_model():
    repo = ModelRepository()
    repo.update(DummyOpenAIEncoderModel(name="openai-embedding-model"))

    actual = repo.get_model("openai-embedding-model")

    assert actual is not None
    assert isinstance(actual, OpenAIEncoderModel)
    assert actual.name == "openai-embedding-model"


def test_adding_rerank_model():
    repo = ModelRepository()
    repo.update(DummyOpenAIEncoderModel(name="rerank-model"))

    actual = repo.get_model("rerank-model")

    assert actual is not None
    assert isinstance(actual, OpenAIEncoderModel)
    assert actual.name == "rerank-model"


@pytest.mark.asyncio
async def test_is_model_ready_nonexistent_model():
    repo = ModelRepository()
    actual = await repo.is_model_ready("none-model")
    assert actual is False


@pytest.mark.asyncio
async def test_is_model_ready_kserve_model():
    repo = ModelRepository()
    model = Model(name="kserve-model")
    repo.update(model)
    with patch.object(model, "healthy"):
        model.healthy.side_effect = lambda: model.ready
        actual = await repo.is_model_ready("kserve-model")
        assert actual is False
        model.load()
        actual = await repo.is_model_ready("kserve-model")
        assert actual is True
        assert len(model.healthy.call_args) == 2


@pytest.mark.asyncio
async def test_is_model_ready_openai_model():
    repo = ModelRepository()
    model = DummyOpenAIGenerativeModel(name="openai-model")
    repo.update(model)

    actual = await repo.is_model_ready("openai-model")
    assert actual is True



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_inference_service_list import (
    V1beta1InferenceServiceList,
)  # noqa: E501
from kserve.rest import ApiException


class TestV1beta1InferenceServiceList(unittest.TestCase):
    """V1beta1InferenceServiceList unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1InferenceServiceList
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1beta1_inference_service_list.V1beta1InferenceServiceList()  # noqa: E501
        if include_optional:
            return V1beta1InferenceServiceList(
                api_version="0",
                items=[
                    kserve.models.v1beta1_inference_service.V1beta1InferenceService(
                        api_version="0",
                        kind="0",
                        metadata=None,
                        spec=kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec(
                            explainer=kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                                active_deadline_seconds=56,
                                affinity=None,
                                art=kserve.models.v1beta1_art_explainer_spec.V1beta1ARTExplainerSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    type="0",
                                    working_dir="0",
                                ),
                                automount_service_account_token=True,
                                batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                                    max_batch_size=56,
                                    max_latency=56,
                                    timeout=56,
                                ),
                                canary_traffic_percent=56,
                                container_concurrency=56,
                                containers=[None],
                                dns_config=None,
                                dns_policy="0",
                                enable_service_links=True,
                                ephemeral_containers=[None],
                                host_aliases=[None],
                                host_ipc=True,
                                host_network=True,
                                host_pid=True,
                                hostname="0",
                                image_pull_secrets=[None],
                                init_containers=[None],
                                logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                                    mode="0",
                                    url="0",
                                ),
                                max_replicas=56,
                                min_replicas=56,
                                node_name="0",
                                node_selector={"key": "0"},
                                overhead={"key": None},
                                preemption_policy="0",
                                priority=56,
                                priority_class_name="0",
                                readiness_gates=[None],
                                restart_policy="0",
                                runtime_class_name="0",
                                scheduler_name="0",
                                security_context=None,
                                service_account="0",
                                service_account_name="0",
                                share_process_namespace=True,
                                subdomain="0",
                                termination_grace_period_seconds=56,
                                timeout=56,
                                tolerations=[None],
                                topology_spread_constraints=[None],
                                volumes=[None],
                            ),
                            predictor=kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                                active_deadline_seconds=56,
                                affinity=None,
                                automount_service_account_token=True,
                                canary_traffic_percent=56,
                                container_concurrency=56,
                                dns_config=None,
                                dns_policy="0",
                                enable_service_links=True,
                                host_ipc=True,
                                host_network=True,
                                host_pid=True,
                                hostname="0",
                                max_replicas=56,
                                min_replicas=56,
                                node_name="0",
                                onnx=kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                preemption_policy="0",
                                priority=56,
                                priority_class_name="0",
                                pytorch=kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                restart_policy="0",
                                runtime_class_name="0",
                                scheduler_name="0",
                                security_context=None,
                                service_account="0",
                                service_account_name="0",
                                share_process_namespace=True,
                                sklearn=kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                subdomain="0",
                                tensorflow=kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                termination_grace_period_seconds=56,
                                timeout=56,
                                triton=kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                xgboost=kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                            ),
                            transformer=kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                                active_deadline_seconds=56,
                                affinity=None,
                                automount_service_account_token=True,
                                canary_traffic_percent=56,
                                container_concurrency=56,
                                dns_config=None,
                                dns_policy="0",
                                enable_service_links=True,
                                host_ipc=True,
                                host_network=True,
                                host_pid=True,
                                hostname="0",
                                max_replicas=56,
                                min_replicas=56,
                                node_name="0",
                                preemption_policy="0",
                                priority=56,
                                priority_class_name="0",
                                restart_policy="0",
                                runtime_class_name="0",
                                scheduler_name="0",
                                security_context=None,
                                service_account="0",
                                service_account_name="0",
                                share_process_namespace=True,
                                subdomain="0",
                                termination_grace_period_seconds=56,
                                timeout=56,
                            ),
                        ),
                        status=kserve.models.v1beta1_inference_service_status.V1beta1InferenceServiceStatus(
                            address=None,
                            components={
                                "key": kserve.models.v1beta1_component_status_spec.V1beta1ComponentStatusSpec(
                                    address=None,
                                    latest_created_revision="0",
                                    latest_ready_revision="0",
                                    latest_rolledout_revision="0",
                                    previous_rolledout_revision="0",
                                    traffic=list(),
                                    url=None,
                                )
                            },
                            conditions=[None],
                            observed_generation=56,
                            url=None,
                        ),
                    )
                ],
                kind="0",
                metadata=None,
            )
        else:
            return V1beta1InferenceServiceList(
                items=[
                    kserve.models.v1beta1_inference_service.V1beta1InferenceService(
                        api_version="0",
                        kind="0",
                        metadata=None,
                        spec=kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec(
                            explainer=kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                                active_deadline_seconds=56,
                                affinity=None,
                                art=kserve.models.v1beta1_art_explainer_spec.V1beta1ARTExplainerSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    type="0",
                                    working_dir="0",
                                ),
                                automount_service_account_token=True,
                                batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                                    max_batch_size=56,
                                    max_latency=56,
                                    timeout=56,
                                ),
                                canary_traffic_percent=56,
                                container_concurrency=56,
                                containers=[None],
                                dns_config=None,
                                dns_policy="0",
                                enable_service_links=True,
                                ephemeral_containers=[None],
                                host_aliases=[None],
                                host_ipc=True,
                                host_network=True,
                                host_pid=True,
                                hostname="0",
                                image_pull_secrets=[None],
                                init_containers=[None],
                                logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                                    mode="0",
                                    url="0",
                                ),
                                max_replicas=56,
                                min_replicas=56,
                                node_name="0",
                                node_selector={"key": "0"},
                                overhead={"key": None},
                                preemption_policy="0",
                                priority=56,
                                priority_class_name="0",
                                readiness_gates=[None],
                                restart_policy="0",
                                runtime_class_name="0",
                                scheduler_name="0",
                                security_context=None,
                                service_account="0",
                                service_account_name="0",
                                share_process_namespace=True,
                                subdomain="0",
                                termination_grace_period_seconds=56,
                                timeout=56,
                                tolerations=[None],
                                topology_spread_constraints=[None],
                                volumes=[None],
                            ),
                            predictor=kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                                active_deadline_seconds=56,
                                affinity=None,
                                automount_service_account_token=True,
                                canary_traffic_percent=56,
                                container_concurrency=56,
                                dns_config=None,
                                dns_policy="0",
                                enable_service_links=True,
                                host_ipc=True,
                                host_network=True,
                                host_pid=True,
                                hostname="0",
                                max_replicas=56,
                                min_replicas=56,
                                node_name="0",
                                onnx=kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                preemption_policy="0",
                                priority=56,
                                priority_class_name="0",
                                pytorch=kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                restart_policy="0",
                                runtime_class_name="0",
                                scheduler_name="0",
                                security_context=None,
                                service_account="0",
                                service_account_name="0",
                                share_process_namespace=True,
                                sklearn=kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                subdomain="0",
                                tensorflow=kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                termination_grace_period_seconds=56,
                                timeout=56,
                                triton=kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                                xgboost=kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                                    image="0",
                                    image_pull_policy="0",
                                    lifecycle=None,
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    runtime_version="0",
                                    security_context=None,
                                    startup_probe=None,
                                    stdin=True,
                                    stdin_once=True,
                                    storage_uri="0",
                                    termination_message_path="0",
                                    termination_message_policy="0",
                                    tty=True,
                                    working_dir="0",
                                ),
                            ),
                            transformer=kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                                active_deadline_seconds=56,
                                affinity=None,
                                automount_service_account_token=True,
                                canary_traffic_percent=56,
                                container_concurrency=56,
                                dns_config=None,
                                dns_policy="0",
                                enable_service_links=True,
                                host_ipc=True,
                                host_network=True,
                                host_pid=True,
                                hostname="0",
                                max_replicas=56,
                                min_replicas=56,
                                node_name="0",
                                preemption_policy="0",
                                priority=56,
                                priority_class_name="0",
                                restart_policy="0",
                                runtime_class_name="0",
                                scheduler_name="0",
                                security_context=None,
                                service_account="0",
                                service_account_name="0",
                                share_process_namespace=True,
                                subdomain="0",
                                termination_grace_period_seconds=56,
                                timeout=56,
                            ),
                        ),
                        status=kserve.models.v1beta1_inference_service_status.V1beta1InferenceServiceStatus(
                            address=None,
                            components={
                                "key": kserve.models.v1beta1_component_status_spec.V1beta1ComponentStatusSpec(
                                    address=None,
                                    latest_created_revision="0",
                                    latest_ready_revision="0",
                                    latest_rolledout_revision="0",
                                    previous_rolledout_revision="0",
                                    traffic=list(),
                                    url=None,
                                )
                            },
                            conditions=[None],
                            observed_generation=56,
                            url=None,
                        ),
                    )
                ],
            )

    def testV1beta1InferenceServiceList(self):
        """Test V1beta1InferenceServiceList"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1beta1PodMetrics(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'backend': 'str',
        'metric_names': 'list[str]',
        'operation_over_time': 'str',
        'query': 'str',
        'server_address': 'str'
    }

    attribute_map = {
        'backend': 'backend',
        'metric_names': 'metricNames',
        'operation_over_time': 'operationOverTime',
        'query': 'query',
        'server_address': 'serverAddress'
    }

    def __init__(self, backend='', metric_names=None, operation_over_time=None, query=None, server_address=None, local_vars_configuration=None):  # noqa: E501
        """V1beta1PodMetrics - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._backend = None
        self._metric_names = None
        self._operation_over_time = None
        self._query = None
        self._server_address = None
        self.discriminator = None

        if backend is not None:
            self.backend = backend
        if metric_names is not None:
            self.metric_names = metric_names
        if operation_over_time is not None:
            self.operation_over_time = operation_over_time
        if query is not None:
            self.query = query
        if server_address is not None:
            self.server_address = server_address

    @property
    def backend(self):
        """Gets the backend of this V1beta1PodMetrics.  # noqa: E501

        Backend defines the scaling metric type watched by the autoscaler. Possible value: opentelemetry.  # noqa: E501

        :return: The backend of this V1beta1PodMetrics.  # noqa: E501
        :rtype: str
        """
        return self._backend

    @backend.setter
    def backend(self, backend):
        """Sets the backend of this V1beta1PodMetrics.

        Backend defines the scaling metric type watched by the autoscaler. Possible value: opentelemetry.  # noqa: E501

        :param backend: The backend of this V1beta1PodMetrics.  # noqa: E501
        :type: str
        """

        self._backend = backend

    @property
    def metric_names(self):
        """Gets the metric_names of this V1beta1PodMetrics.  # noqa: E501

        MetricNames is the list of metric names in the backend.  # noqa: E501

        :return: The metric_names of this V1beta1PodMetrics.  # noqa: E501
        :rtype: list[str]
        """
        return self._metric_names

    @metric_names.setter
    def metric_names(self, metric_names):
        """Sets the metric_names of this V1beta1PodMetrics.

        MetricNames is the list of metric names in the backend.  # noqa: E501

        :param metric_names: The metric_names of this V1beta1PodMetrics.  # noqa: E501
        :type: list[str]
        """

        self._metric_names = metric_names

    @property
    def operation_over_time(self):
        """Gets the operation_over_time of this V1beta1PodMetrics.  # noqa: E501

        OperationOverTime specifies the operation to aggregate the metrics over time. Possible values are last_one, avg, max, min, rate, count. Default is 'last_one'.  # noqa: E501

        :return: The operation_over_time of this V1beta1PodMetrics.  # noqa: E501
        :rtype: str
        """
        return self._operation_over_time

    @operation_over_time.setter
    def operation_over_time(self, operation_over_time):
        """Sets the operation_over_time of this V1beta1PodMetrics.

        OperationOverTime specifies the operation to aggregate the metrics over time. Possible values are last_one, avg, max, min, rate, count. Default is 'last_one'.  # noqa: E501

        :param operation_over_time: The operation_over_time of this V1beta1PodMetrics.  # noqa: E501
        :type: str
        """

        self._operation_over_time = operation_over_time

    @property
    def query(self):
        """Gets the query of this V1beta1PodMetrics.  # noqa: E501

        Query specifies the query to run to get metrics from the PodsMetricsBackend.  # noqa: E501

        :return: The query of this V1beta1PodMetrics.  # noqa: E501
        :rtype: str
        """
        return self._query

    @query.setter
    def query(self, query):
        """Sets the query of this V1beta1PodMetrics.

        Query specifies the query to run to get metrics from the PodsMetricsBackend.  # noqa: E501

        :param query: The query of this V1beta1PodMetrics.  # noqa: E501
        :type: str
        """

        self._query = query

    @property
    def server_address(self):
        """Gets the server_address of this V1beta1PodMetrics.  # noqa: E501

        ServerAddress specifies the address of the PodsMetricsBackend server.  # noqa: E501

        :return: The server_address of this V1beta1PodMetrics.  # noqa: E501
        :rtype: str
        """
        return self._server_address

    @server_address.setter
    def server_address(self, server_address):
        """Sets the server_address of this V1beta1PodMetrics.

        ServerAddress specifies the address of the PodsMetricsBackend server.  # noqa: E501

        :param server_address: The server_address of this V1beta1PodMetrics.  # noqa: E501
        :type: str
        """

        self._server_address = server_address

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1beta1PodMetrics):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1beta1PodMetrics):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict, Optional

from .model import BaseKServeModel

MODEL_MOUNT_DIRS = "/mnt/models"


class ModelRepository:
    """Model repository interface.

    It follows NVIDIA Triton's `Model Repository Extension`_.

    .. _Model Repository Extension:
        https://github.com/triton-inference-server/server/blob/main/docs/protocol/extension_model_repository.md
    """

    def __init__(self, models_dir: str = MODEL_MOUNT_DIRS):
        self.models: Dict[str, BaseKServeModel] = {}
        self.models_dir = models_dir

    def load_models(self):
        for name in os.listdir(self.models_dir):
            d = os.path.join(self.models_dir, name)
            if os.path.isdir(d):
                self.load_model(name)

    def set_models_dir(self, models_dir):  # used for unit tests
        self.models_dir = models_dir

    def get_model(self, name: str) -> Optional[BaseKServeModel]:
        return self.models.get(name, None)

    def get_models(self) -> Dict[str, BaseKServeModel]:
        return self.models

    async def is_model_ready(self, name: str):
        model = self.get_model(name)
        if not model:
            return False
        return await model.healthy()

    def update(self, model: BaseKServeModel, name: Optional[str] = None):
        """
        Update or add a model to the repository.
        Args:
            model (BaseKServeModel): The model to be added or updated in the repository.
            name (Optional[str], optional): The name to use for the model.
                If not provided, the model's own name attribute will be used. Defaults to None.
                This can be used to provide additional names for the same model.
        Returns:
            None
        """

        if name:
            self.models[name] = model
        else:
            self.models[model.name] = model

    def load(self, name: str) -> bool:
        pass

    def load_model(self, name: str) -> bool:
        pass

    def unload(self, name: str):
        if name in self.models:
            model = self.models[name]
            if callable(getattr(model, "stop", None)):
                model.stop()
            if model.engine:
                model.stop_engine()
            del self.models[name]
        else:
            raise KeyError(f"model with name {name} does not exist")



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration classes for KServe."""

from .constants.constants import PredictorProtocol


class PredictorConfig:
    def __init__(
        self,
        predictor_host: str,
        predictor_protocol: str = PredictorProtocol.REST_V1.value,
        predictor_use_ssl: bool = False,
        predictor_request_timeout_seconds: int = 600,
        predictor_request_retries: int = 0,
        predictor_health_check: bool = False,
    ):
        """The configuration for the http call to the predictor

        Args:
            predictor_host: The host name of the predictor
            predictor_protocol: The inference protocol used for predictor http call
            predictor_use_ssl: Enable using ssl for http connection to the predictor
            predictor_request_timeout_seconds: The request timeout seconds for the predictor http call. Default is 600 seconds.
            predictor_request_retries: The number of retries if the predictor request fails. Default is 0.
            predictor_health_check: Enable predictor health check
        """
        self._predictor_host = predictor_host
        self._predictor_protocol = predictor_protocol
        self._predictor_use_ssl = predictor_use_ssl
        self._predictor_request_timeout_seconds = predictor_request_timeout_seconds
        self._predictor_request_retries = predictor_request_retries
        self._predictor_health_check = predictor_health_check

    @property
    def predictor_host(self) -> str:
        """Get the predictor host."""
        return self._predictor_host

    @property
    def predictor_protocol(self) -> str:
        """Get the predictor protocol."""
        return self._predictor_protocol

    @property
    def predictor_use_ssl(self) -> bool:
        """Get the predictor use ssl flag."""
        return self._predictor_use_ssl

    @property
    def predictor_request_timeout_seconds(self) -> int:
        """Get the predictor request timeout in seconds."""
        return self._predictor_request_timeout_seconds

    @property
    def predictor_request_retries(self) -> int:
        """Get the predictor request retries."""
        return self._predictor_request_retries

    @property
    def predictor_health_check(self) -> bool:
        """Get the predictor health check flag."""
        return self._predictor_health_check

    @property
    def predictor_base_url(self) -> str:
        """
        Get the base url for the predictor.

        Returns:
            str: The base url for the predictor
        """
        protocol = "https" if self._predictor_use_ssl else "http"
        return f"{protocol}://{self._predictor_host}"

    @property
    def protocol(self) -> str:
        """Alias for predictor_protocol for backward compatibility."""
        return self._predictor_protocol

    @property
    def timeout(self) -> int:
        """Alias for predictor_request_timeout_seconds for backward compatibility."""
        return self._predictor_request_timeout_seconds

    @property
    def retries(self) -> int:
        """Alias for predictor_request_retries for backward compatibility."""
        return self._predictor_request_retries

    @property
    def use_ssl(self) -> bool:
        """Alias for predictor_use_ssl for backward compatibility."""
        return self._predictor_use_ssl
